# quick and dirty deduplicate of agent with duplicated name within an authority
for agent in rql('DISTINCT Any X WHERE X is Agent, X name XN, X authority A, '
                 'Y is Agent, Y name XN, Y authority A, '
                 'Y eid YE, X eid >YE').entities():
    agent.cw_delete()
commit()

for ertype in ('generated', 'used', 'associated_with', 'place_address', 'new_version_of',
               'agent_kind', 'authority_record', 'use_profile',
               'OrganizationUnit', 'Agent', 'ArkNameAssigningAuthority'):
    sync_schema_props_perms(ertype)


sql("DELETE FROM container_relation WHERE EXISTS("
    "SELECT FROM entities WHERE eid_to=eid AND "
    "type IN ('AuthorityRecord', 'ConceptScheme', 'Concept'))")
commit()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref specific hooks and operations"""

from datetime import datetime

import psycopg2

from logilab.common.registry import objectify_predicate

from yams import ValidationError
from yams.schema import role_name

from cubicweb import _
from cubicweb.schema import META_RTYPES
from cubicweb.predicates import (is_instance, adaptable, on_fire_transition,
                                 relation_possible)
from cubicweb.server import hook
from cubicweb.hooks import metadata

from cubicweb_skos.hooks import ReplaceExternalUriByEntityHook
from .ark import (
    match as ark_match,
    insert_ark,
)


# skip relations involved in the logging itself and some others that should not
# be logged nor impact container's modification date
IGNORE_RELATIONS = set(('used', 'generated', 'associated_with',
                        'primary_email',
                        'container',
                        'use_authorityrecord', 'related_concept_scheme', 'use_profile',
                        'new_version_of', 'clone_of',
                        'scheme_entity_type', 'scheme_relation_type'))


@objectify_predicate
def contained_relation(cls, req, rtype, eidfrom, eidto, **kwargs):
    """Predicate that returns True for relation from/to a container or contained entity.
    """
    if rtype in IGNORE_RELATIONS:
        return 0
    for eid in (eidfrom, eidto):
        entity = req.entity_from_eid(eid)
        for interface in ('IContainer', 'IContained'):
            if entity.cw_adapt_to(interface):
                # skip meta relation of contained entities, only consider them on the container
                if interface == 'IContained' and rtype in META_RTYPES:
                    return 0
                return 1
    return 0


def qualify_relation(subj, rtype, obj):
    """Return a list of (relationship, role, target) qualifying the relation.

    Options are:

    * `relationship` = 'parent' and `target` the parent entity adapter
    * `relationship` = 'border' and `target` the container or contained adapter

    In both cases `role` is the role of the target in the relation.
    """
    # first attempt to get the more specific adapter for subject and object of the relation
    for interface in ('IContainer', 'IContained'):
        s_adapter = subj.cw_adapt_to(interface)
        if s_adapter is not None:
            break
    for interface in ('IContainer', 'IContained'):
        o_adapter = obj.cw_adapt_to(interface)
        if o_adapter is not None:
            break
    # then check if the relation is a parent relation
    if (s_adapter is not None and s_adapter.__regid__ == 'IContained'
            and (rtype, 'subject') in s_adapter.parent_relations):
        if o_adapter is None:
            # relation actually outside a container (e.g. EmailAddress use_email CWUser)
            return []
        return [('parent', 'object', o_adapter)]
    if (o_adapter is not None and o_adapter.__regid__ == 'IContained'
            and (rtype, 'object') in o_adapter.parent_relations):
        if s_adapter is None:
            # relation actually outside a container (e.g. EmailAddress use_email CWUser)
            return []
        return [('parent', 'subject', s_adapter)]
    # from here, we can assume rtype is not a parentship relation. Notice the relation may concerns
    # two distinct containers
    assert s_adapter or o_adapter
    result = []
    # if relation concerns two containers and one of the entity is a concept,
    # don't log action on the concept
    if s_adapter is not None and (o_adapter is None or _log_relation(rtype, subj, obj)):
        result.append(('border', 'subject', s_adapter))
    if o_adapter is not None and (s_adapter is None or _log_relation(rtype, obj, subj)):
        result.append(('border', 'object', o_adapter))
    return result


def _log_relation(rtype, entity, target):
    """Return boolean indicating wether the relation modification should be logged.

    This is called when relation's ends are in different containers, with
    `entity` as the potentiel activity holder and `target` the other relation's
    end.
    """
    # code_keyword_type is qualifying the subject scheme, don't log the
    # modification on the target concept
    if rtype == 'code_keyword_type':
        return entity.cw_etype == 'ConceptScheme'
    # if entity is a scheme or concept, only log if the other ends is also a
    # concept or scheme
    if entity.cw_etype in ('Concept', 'ConceptScheme'):
        return target.cw_etype in ('Concept', 'ConceptScheme')
    return True


# generic hooks to record operations if something changed in a compound tree ###

class AddOrRemoveChildrenHook(hook.Hook):
    """Some relation involved in a compound graph is added or removed."""
    __regid__ = 'compound.graph.updated'
    __select__ = hook.Hook.__select__ & contained_relation()
    events = ('before_add_relation', 'before_delete_relation')
    category = 'metadata'

    def __call__(self):
        for relationship, role, target in qualify_relation(self._cw.entity_from_eid(self.eidfrom),
                                                           self.rtype,
                                                           self._cw.entity_from_eid(self.eidto)):
            entity = target.entity
            UpdateModificationDateOp.get_instance(self._cw).add_entity(entity)


class UpdateEntityHook(hook.Hook):
    """Some entity involved in a compound graph is updated."""
    __regid__ = 'compound.graph.updated'
    __select__ = hook.Hook.__select__ & adaptable('IContainer', 'IContained')
    events = ('before_update_entity',)
    category = 'metadata'

    def __call__(self):
        UpdateModificationDateOp.get_instance(self._cw).add_entity(self.entity)


# Auto-update of modification date in a compound tree ##########################

class UpdateModificationDateOp(hook.DataOperationMixIn, hook.Operation):
    """Data operation updating the modification date of its data entities."""

    def precommit_event(self):
        cnx = self.cnx
        now = datetime.utcnow()
        with cnx.deny_all_hooks_but():
            for eid in self.get_data():
                if cnx.deleted_in_transaction(eid) or cnx.added_in_transaction(eid):
                    continue
                entity = cnx.entity_from_eid(eid)
                entity.cw_set(modification_date=now)

    def add_entity(self, entity):
        """Add entity, its parent entities (up to the container root) for update of their
        modification date at commit time.
        """
        self.add_data(entity.eid)
        safety_belt = set((entity.eid,))
        while True:
            contained = entity.cw_adapt_to('IContained')
            if contained is None:
                assert entity.cw_adapt_to('IContainer')
                break
            else:
                entity = contained.parent
                if entity is None:
                    break
                if entity.eid in safety_belt:
                    self.warning('loop detected implying %s(%s)', entity.cw_etype, entity.eid)
                    break
                self.add_data(entity.eid)
                safety_belt.add(entity.eid)


# Transformation of ExternalUri to AuthorityRecord

class ReplaceExternalUriByAuthorityRecordHook(ReplaceExternalUriByEntityHook):
    """Replace ExternalUri by an AuthorityRecord"""
    __select__ = ReplaceExternalUriByEntityHook.__select__ & is_instance('AuthorityRecord')


# ARK generation ###############################################################

class AssignARKHook(hook.Hook):
    """When an entity supporting ARK is created, assign it a local ark to."""
    __regid__ = 'saem.ark.assign'
    __select__ = hook.Hook.__select__ & relation_possible('ark')
    events = ('before_add_entity',)
    order = metadata.InitMetaAttrsHook.order - 1
    category = 'metadata'

    def __call__(self):
        if not self.entity.cw_edited.get('ark'):
            naa_what = self.entity.cw_adapt_to('IArkNAALocator').naa_what()
        else:
            naa_what = None
        set_ark_and_cwuri(self._cw, self.entity.cw_edited, naa_what=naa_what)


def set_ark_and_cwuri(cw, entity_attrs, naa_what=None):
    ark = entity_attrs.get('ark')
    if ark:
        m = ark_match(ark, external=True)
        msg = _('malformatted ARK idenfitier %(ark)s (expecting [ark:/]NAAN/Name[Qualifier])')
        msgargs = {'ark': ark}
        if m is None:
            raise ValidationError(None, {'ark': msg}, msgargs=msgargs)
        matched = m.groupdict()
        naan, name = matched.pop('naan'), matched.pop('name')
        qualifier = matched.get('qualifier')
        try:
            ark = insert_ark(cw, naan, name, qualifier)
        except psycopg2.IntegrityError:
            raise ValidationError(
                None, {'ark': _('%(ark)s already exists')}, msgargs={'ark': ark})
    else:
        cwuri = entity_attrs.get('cwuri')
        ark = None if cwuri is None else extract_ark(cwuri)
        if ark is None:
            if naa_what is None:
                msg = _('an ARK identifier has to be generated but no Name Assigning Authority is '
                        'specified')
                raise ValidationError(None, {None: msg})
            generator = cw.vreg['adapters'].select(
                'IARKGenerator', cw,
                naa_what=naa_what, **entity_attrs)
            ark = generator.generate_ark()
    entity_attrs['ark'] = ark
    if 'cwuri' not in entity_attrs:
        # store ark as cwuri, not an URL so it's easier to move the database while still easy to get
        # an URL from there (see the cwuri_url function) XXX (syt) any other reason to do so? there
        # is also probably some constraint on (re)import or something similar but I don't recall
        # right now
        entity_attrs['cwuri'] = u'ark:/' + ark


def extract_ark(url):
    """Extract ARK identifier from an URL, return it or None if not found or malformed.
    """
    try:
        _, ark = url.split('ark:/', 1)
    except ValueError:
        return None
    parts = ark.split('/')
    if len(parts) < 2:
        return None
    ark = '/'.join(parts[:2])
    for delim in '#?':
        ark = ark.split(delim, 1)[0]
    return ark


# Life-cycle logging ###########################################################

class Record(object):
    """Temporary representation of some activities to be recorded, for accumulation per entity
    prior to merge in an operation.
    """
    def __init__(self, events):
        if not isinstance(events, set):
            events = set(events)
        self.events = events

    def __str__(self):
        return ','.join(sorted('%s %s' % ev for ev in self.events))

    def merge(self, other):
        """Merge the activity with another."""
        self.events |= other.events

    def as_dict(self, _):
        """Return a dictionary suitable for creation of the Activity entity."""
        now = datetime.utcnow()
        record = {'start': now, 'end': now}
        # merge events on the same target
        targets_per_event_type = {}
        events_by_target = {}
        for ev_type, ev_target in self.events:
            events_by_target.setdefault(ev_target, set([])).add(ev_type)
        for target, target_events in events_by_target.items():
            # if there are several event types for the same target (e.g. added and removed a
            # relation), group them into the "modified" event type
            if len(target_events) > 1:
                ev_type = 'modified'
            else:
                ev_type = next(iter(target_events))
            targets_per_event_type.setdefault(_(ev_type), []).append(target)
        # now generate proper messages for each event type
        msgs = []
        # iterates on event types to get expected order
        for ev_type in (_('created'), _('modified'), _('added'), _('removed')):
            if ev_type not in targets_per_event_type:
                continue
            modified = sorted(_(x).lower() for x in targets_per_event_type[ev_type])
            msgs.append(u'%s %s' % (ev_type, u', '.join(modified)))
        if len(msgs) == 1:
            record['description'] = msgs[0]
        else:
            record['description'] = u'\n'.join(u'* ' + msg for msg in msgs)
        # complete record with other metadata then we're done
        record['description_format'] = u'text/rest'
        record['type'] = u'create' if _('created') in targets_per_event_type else u'modify'
        return record


class AddActivityOperation(hook.DataOperationMixIn, hook.LateOperation):
    """The operation responsible to merge activites to their container then to record them."""
    def precommit_event(self):
        # translate using site default language
        lang = self.cnx.vreg.property_value('ui.language')
        _ = self.cnx.vreg.config.translations[lang][0]
        # first, merge all activities (there may be several for the same container entity)
        activity_per_container = {}
        for entity, activity in self.get_data():
            # find the entity on which we should log the activity
            icontainer = entity.cw_adapt_to('IContainer')
            if icontainer is None:
                icontainer = entity.cw_adapt_to('IContained')
                if icontainer is None:
                    # "free" entity, no logging
                    self.warning("%s is not in a container, don't record activity", entity)
                    continue
            container = icontainer.container
            # if the container is not yet linked or has been deleted in this transaction, we've
            # nothing to log on
            if container is None or self.cnx.deleted_in_transaction(container.eid):
                continue
            # if the container has been created in this transaction, discard all activities but the
            # creation
            if self.cnx.added_in_transaction(container.eid):
                if ('created', entity.cw_etype) in activity.events:
                    # only keep the creation event
                    assert container not in activity_per_container
                    activity_per_container[container] = activity
            else:
                try:
                    activity_per_container[container].merge(activity)
                except KeyError:
                    activity_per_container[container] = activity
        # who?
        if self.cnx.user.eid == -1:  # internal manager
            user = None
        else:
            user = self.cnx.user
        # then record the activities
        for container, activity in activity_per_container.items():
            kwargs = activity.as_dict(_)
            if user is not None:
                kwargs.setdefault('associated_with', user)
            adapted = container.cw_adapt_to('IRecordable')
            if adapted is not None:
                adapted.add_activity(**kwargs)


class LogContainerCreationHook(hook.Hook):
    """Add an Activity upon creation of a container entity (`IContainer`). There is no need for a
    hook on `IContained` creation, we'll catch creation of the relation linking them to their
    container.
    """
    __select__ = hook.Hook.__select__ & adaptable('IContainer')
    __regid__ = 'saem.log.add'
    events = ('after_add_entity',)
    category = 'prov.logging'

    def __call__(self):
        activity = Record([('created', self.entity.cw_etype)])
        AddActivityOperation.get_instance(self._cw).add_data((self.entity, activity))


class LogModificationHook(hook.Hook):
    """Add an Activity upon modification of a container (`IContainer`) or contained (`IContained`)
    entity. The message should be adapted accordingly.
    """
    __select__ = hook.Hook.__select__ & adaptable('IContainer', 'IContained')
    __regid__ = 'saem.log.update'
    events = ('after_update_entity',)
    category = 'prov.logging'

    def __call__(self):
        entity = self.entity
        if entity.cw_adapt_to('IContainer'):
            attrs = set(attr for attr in entity.cw_edited if attr not in META_RTYPES)
            if not attrs:
                return
            events = [('modified', attr) for attr in attrs]
        else:  # IContained
            parent_relation = entity.cw_adapt_to('IContained').parent_relation()
            if parent_relation is None:
                # not yet bound to a parent, skip this activity (later addition of the relation will
                # trigger the record if necessary)
                return
            rtype, role = parent_relation
            # role is the role of the entity, we want to qualify the relation by the role of the
            # parent
            if role == 'subject':
                rtype += '_object'
            events = [('modified', rtype)]
        activity = Record(events)
        AddActivityOperation.get_instance(self._cw).add_data((entity, activity))


class LogRelationHook(hook.Hook):
    """Add an Activity upon modification of a container (`IContainer`) or contained (`IContained`)
    entity. The message should be adapted accordingly.
    """
    __select__ = hook.Hook.__select__ & contained_relation()
    __regid__ = 'saem.log.relation'
    events = ('after_add_relation', 'before_delete_relation')
    category = 'prov.logging'

    def __call__(self):
        for relationship, role, target in qualify_relation(self._cw.entity_from_eid(self.eidfrom),
                                                           self.rtype,
                                                           self._cw.entity_from_eid(self.eidto)):
            rtype = self.rtype
            if role == 'object':
                rtype += '_object'
            if self.event == 'after_add_relation':
                events = [('added', rtype)]
                self.add_activity(target.entity, events)
            else:  # before_delete_relation
                events = [('removed', rtype)]
                # in the case of a relation deletion, we've to find the container right away without
                # waiting for the operation since the parent relation may have been removed at this
                # point
                container = target.container
                if container is not None:
                    self.add_activity(container, events)

    def add_activity(self, entity, events):
        activity = Record(events)
        AddActivityOperation.get_instance(self._cw).add_data((entity, activity))


class DepreciateProfileOnReplacePublishedHook(hook.Hook):
    """Depreciate a SEDA profile once the new version is published."""
    __select__ = hook.Hook.__select__ & on_fire_transition('SEDAArchiveTransfer', 'publish')
    __regid__ = 'seda.deprecate_profile_on_replace_published'
    events = ('after_add_entity',)

    def __call__(self):
        replaced = self.entity.for_entity.new_version_of
        if replaced:
            wf = replaced[0].cw_adapt_to('IWorkflowable')
            wf.fire_transition_if_possible("deprecate")


class UnlinkDeprecatedProfileFromOrganizationUnit(hook.Hook):
    """Remove 'use_profile' relation of a SEDAArchiveTransfer when it gets
    deprecated.
    """
    __select__ = (
        hook.Hook.__select__
        & on_fire_transition('SEDAArchiveTransfer', 'deprecate')
    )
    __regid__ = 'seda.unlink_deprecated_profile_from_organization_unit'
    events = ('after_add_entity',)

    def __call__(self):
        self._cw.execute('DELETE O use_profile X WHERE X eid %(x)s',
                         {'x': self.entity.for_entity.eid})


class CheckSchemeRelationTypeBeforeDeprecation(hook.Hook):
    """Prevent ConceptScheme deprecation if it used as a "scheme_relation_type".
    """
    __select__ = (
        hook.Hook.__select__
        & on_fire_transition('ConceptScheme', 'deprecate')
    )
    __regid__ = 'saem.check-scheme-relation-type-before-deprecation'
    events = ('after_add_entity',)

    def __call__(self):
        scheme_to_seda_rtypes = [
            ('scheme_relation_type', 'subject'),
            ('scheme_entity_type', 'subject'),
            ('seda_keyword_reference_to_scheme', 'object'),
            ('seda_message_digest_algorithm_code_list_version_to', 'object'),
            ('seda_mime_type_code_list_version_to', 'object'),
            ('seda_encoding_code_list_version_to', 'object'),
            ('seda_file_format_code_list_version_to', 'object'),
            ('seda_compression_algorithm_code_list_version_to', 'object'),
            ('seda_data_object_version_code_list_version_to', 'object'),
            ('seda_storage_rule_code_list_version_to', 'object'),
            ('seda_appraisal_rule_code_list_version_to', 'object'),
            ('seda_access_rule_code_list_version_to', 'object'),
            ('seda_dissemination_rule_code_list_version_to', 'object'),
            ('seda_reuse_rule_code_list_version_to', 'object'),
            ('seda_classification_rule_code_list_version_to', 'object'),
            ('seda_acquisition_information_code_list_version_to', 'object'),
            ('seda_relationship_code_list_version_to', 'object'),
        ]
        qs = ' OR '.join(
            'EXISTS({subject} {rtype} {object})'.format(
                subject='X' if role == 'subject' else 'Y%d' % idx,
                object='X' if role == 'object' else 'Y%d' % idx,
                rtype=rtype,
            )
            for idx, (rtype, role) in enumerate(scheme_to_seda_rtypes)
        )
        rset = self._cw.execute('Any X WHERE X eid %(x)s, ' + qs,
                                {'x': self.entity.for_entity.eid})
        if rset:
            raise ValidationError(
                self.entity.for_entity.eid,
                {None: _('this concept scheme is used in at least one SEDA constraint')},
            )


# Relations deposit agent - archival agent  ####################################

class DontDeleteUnitIfArchival(hook.Hook):
    """Before deleting an agent, if it is an archival agent, make sure it isn't
    associated to an organization through the "archival_unit" relation.
    """
    __select__ = hook.Hook.__select__ & hook.match_rtype('archival_role')
    __regid__ = 'saem.archival_role.delete'
    events = ('before_delete_relation', )

    def __call__(self):
        ou = self._cw.entity_from_eid(self.eidfrom)
        role = self._cw.entity_from_eid(self.eidto)
        if role.name == u'archival' and ou.reverse_archival_unit:
            errors = {role_name('archival_role', 'subject'):
                      _('this organization unit is the archival unit of some organization, '
                        'therefore the role "archival" cannot be deleted')}
            raise ValidationError(ou.eid, errors)


class AuthorityRecordUsedBySetDefault(hook.Hook):
    """Set a default value for `OrganizationUnit use_authorityrecord
    AuthorityRecord` upon addition of AuthorityRecord entity.
    """
    __regid__ = 'saem.set-default-use_authorityrecord-relation'
    __select__ = (hook.Hook.__select__
                  & hook.match_rtype('created_by',
                                     frometypes=('AuthorityRecord',)))
    events = ('after_add_relation', )

    def __call__(self):
        self._cw.execute(
            'SET OU use_authorityrecord AR WHERE AR created_by U, U eid %(u)s, '
            ' AR eid %(ar)s, U authority O, O archival_unit OU',
            {'u': self.eidto, 'ar': self.eidfrom})


# Define index for sorted composite relation ###################################

class InitializeIndexOnCreation(hook.Hook):
    """An entity which is part of a collection is created"""
    __regid__ = 'saem.sortable.created'
    __select__ = hook.Hook.__select__ & adaptable('ISortable')
    events = ('before_add_entity',)

    def __call__(self):
        # Ensure we use an index greater than every other
        self.entity.cw_edited['index'] = self.entity.eid


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)

    from cubicweb.server import ON_COMMIT_ADD_RELATIONS
    from cubicweb_compound import utils
    from . import ConceptSchemeGraph

    # Add relations involved in a composite graph with security setup to "on
    # commit" check step.
    graph = ConceptSchemeGraph(vreg.schema)
    for rdef, __ in utils.mandatory_rdefs(vreg.schema, graph.parent_structure('ConceptScheme')):
        ON_COMMIT_ADD_RELATIONS.add(rdef.rtype)
    ON_COMMIT_ADD_RELATIONS.add('code_keyword_type')



from cubicweb_seda.dataimport import LCSV_FILES
from cubicweb_saem_ref import workflows

for etype in ('SEDAProfile', 'ProfileDocument', 'ProfileArchiveObject'):
    rql('DELETE %s X' % etype)
    commit()

for etype in schema.entities():
    if etype.type.startswith('SEDA') or etype.type in ('ProfileDocument', 'ProfileArchiveObject'):
        drop_entity_type(str(etype))

# delete preexisting concept schemes with the same name: they have been refactored in seda cube
# (added seda 1/seda 2 labels, etc)
for data_descr in LCSV_FILES:
    title = data_descr[0]
    rset = cnx.find('ConceptScheme', title=title)
    if rset:
        rset.one().cw_delete()
        commit()

add_cube('seda')

workflows.define_publication_workflow(add_workflow, 'SEDAArchiveTransfer')

add_cube('eac')



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref server objects"""

from cubicweb.predicates import match_user_groups, match_kwargs
from cubicweb.server import Service
from cubicweb.dataimport.stores import MetadataGenerator

from cubicweb_eac import sobjects as eac

from .. import user_has_naa
from ..hooks import set_ark_and_cwuri


class SAEMMetadataGenerator(MetadataGenerator):
    """SAEM specific metadata generator, necessary when no hooks are activated (e.g. skos import) to
    handle ARK generation and setting of the initial state.
    """

    def __init__(self, cnx, baseurl=None, source=None, naa_what=None):
        super(SAEMMetadataGenerator, self).__init__(cnx, baseurl, source)
        if naa_what is None and source is not None:
            source_entity = cnx.find('CWSource', eid=source.eid).one()
            if source_entity.reverse_through_cw_source:
                skos_source = source_entity.reverse_through_cw_source[0]
                if skos_source.ark_naa:
                    naa_what = skos_source.ark_naa[0].what
        self.naa_what = naa_what

    def etype_rels(self, etype):
        etype_rels = super(SAEMMetadataGenerator, self).etype_rels(etype)
        if etype == 'ConceptScheme':
            etype_rels.append('in_state')
        return etype_rels

    def gen_in_state(self, etype):
        wf_state = self._cnx.execute('Any S WHERE ET default_workflow WF, ET name %(etype)s, '
                                     'WF initial_state S', {'etype': etype}).one()
        return wf_state.eid

    def entity_attrs(self, etype):
        entity_attrs = super(SAEMMetadataGenerator, self).entity_attrs(etype)
        if etype in ('ConceptScheme', 'Concept', 'AuthorityRecord', 'Agent'):
            entity_attrs.insert(0, 'ark')  # insert before cwuri
        return entity_attrs

    def gen_ark(self, etype, eid, attrs):
        """ARK generation callback"""
        set_ark_and_cwuri(self._cnx, attrs, naa_what=self.naa_what)
        # returning value, even if set by the function, is necessary due to the meta generator
        # implementation
        return attrs['ark']


class EACImportService(eac.EACImportService):
    """Service to import an AuthorityRecord from an EAC XML file - overriden from the EAC cube to
    adapt selector and automatically set ark_naa / authority relations.
    """

    __select__ = eac.EACImportService.__select__ & (match_kwargs('authority') | user_has_naa())

    def import_eac_stream(self, stream, import_log, store, extid2eid=None, naa=None, **kwargs):
        authority = self.cw_extra_kwargs.get('authority')
        if authority is None:
            authority = self._cw.user.authority[0]
        if naa is None:
            naa = authority.ark_naa[0]
        self._authority = authority
        self._naa = naa
        return super(EACImportService, self).import_eac_stream(stream, import_log, store,
                                                               extid2eid=extid2eid, **kwargs)

    def external_entities_stream(self, extentities, extid2eid):
        extentities = super(EACImportService, self).external_entities_stream(extentities, extid2eid)

        extid2eid[self._naa.cwuri] = self._naa.eid
        extid2eid[self._authority.cwuri] = self._authority.eid

        def set_authority_or_naa(extentity):
            """insert function to set parent authority in the ext-entities stream"""
            if extentity.etype == 'AuthorityRecord':
                extentity.values['ark_naa'] = set([self._naa.cwuri])
            elif extentity.etype == 'Agent':
                extentity.values['authority'] = set([self._authority.cwuri])
            return extentity

        return map(set_authority_or_naa, extentities)


class AllocateArk(Service):
    """Service to allocate an ark identifier given an
    ArkNameAssigningAuthority entity.
    """
    __regid__ = 'saem.attribute-ark'
    __select__ = match_user_groups('managers', 'users')

    def call(self, naa):
        generator = self._cw.vreg['adapters'].select('IARKGenerator', self._cw,
                                                     naa_what=naa.what)
        return generator.generate_ark()


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, [EACImportService])
    vreg.register_and_replace(EACImportService, eac.EACImportService)



container_etypes = ('SEDAArchiveTransfer', 'SEDAArchiveUnit')
for otype in container_etypes:
    drop_relation_definition('Activity', 'container', otype)

sql("DELETE FROM container_relation WHERE eid_from IN"
    " (SELECT c.eid_from FROM container_relation AS c"
    "   JOIN entities AS e ON e.eid=c.eid_from WHERE e.type='Activity');")
commit()

sync_schema_props_perms('seda_archive_unit')
sync_schema_props_perms('SEDAArchiveUnit')



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for hooks"""

from datetime import datetime, timedelta
from time import mktime
import unittest

import pytz

from cubicweb import ValidationError
from cubicweb.devtools import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_saem_ref.hooks import extract_ark

import testutils


def setUpModule():
    testutils.startpgcluster(__file__)


YESTERDAY = datetime.now(tz=pytz.utc) - timedelta(days=1)


class SAEMRefHooksTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def assertMDNow(self, entity):
        entity.cw_clear_all_caches()
        self.assertAlmostEqual(mktime(entity.modification_date.timetuple()),
                               mktime(datetime.now(tz=pytz.utc).timetuple()), delta=60)

    def resetMD(self, cnx, *entities):
        for entity in entities:
            with cnx.deny_all_hooks_but():
                entity.cw_set(modification_date=YESTERDAY)
        cnx.commit()

    def test_reset_md(self):
        """Ensure `resetMD` method above works."""
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'bob')
            cnx.commit()
            self.resetMD(cnx, agent)
            agent.cw_clear_all_caches()
            self.assertEqual(agent.modification_date, YESTERDAY)

    def test_sync_scheme_md(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus',
                                       ark_naa=testutils.naa(cnx))
            kt_scheme = testutils.scheme_for_type(cnx, 'seda_keyword_type_to', None, u'theme')
            cnx.commit()
            self.resetMD(cnx, scheme)
            c1 = scheme.add_concept(u'concept 1', language_code=u'fr')
            cnx.commit()
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme)
            lab = cnx.create_entity('Label', label=u'concept 1.1', language_code=u'en',
                                    label_of=c1)
            cnx.commit()
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme)
            lab.cw_set(label=u'concept 1.1.1')
            cnx.commit()
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme, c1)
            c2 = c1.add_concept(u'sub concept')
            cnx.commit()
            self.assertMDNow(c1)
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme, c1, c2)
            c2.add_concept(u'sub-sub concept')
            cnx.commit()
            self.assertMDNow(c2)
            self.assertMDNow(c1)
            self.assertMDNow(scheme)
            self.resetMD(cnx, scheme, c1, c2)
            c2.preferred_label[0].cw_set(label=u'sub concept 2')
            cnx.commit()
            self.assertMDNow(c2)
            self.assertMDNow(c1)
            self.assertMDNow(scheme)

            self.resetMD(cnx, scheme)
            self.resetMD(cnx, kt_scheme)
            cnx.commit()
            scheme.cw_set(code_keyword_type=kt_scheme.reverse_in_scheme[0])
            cnx.commit()
            kt_scheme.cw_clear_all_caches()
            self.assertEqual(kt_scheme.modification_date, YESTERDAY)
            self.assertMDNow(scheme)

    def test_sync_profile_md(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = testutils.scheme_for_type(cnx, 'seda_rule', 'SEDASeqAccessRuleRule', u'AR038')
            transfer = testutils.setup_profile(cnx)
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            cnx.commit()
            transfer.cw_set(modification_date=YESTERDAY)
            cnx.commit()

            # edit composite children
            unit_alt_seq.reverse_seda_title[0].cw_set(title=u'archive name')
            cnx.commit()
            self.assertMDNow(transfer)
            self.resetMD(cnx, transfer)
            self.resetMD(cnx, scheme)
            # edit relation to a composite children
            code = scheme.reverse_in_scheme[0]
            rule_seq = cnx.create_entity('SEDASeqAccessRuleRule',
                                         reverse_seda_start_date=cnx.create_entity('SEDAStartDate'),
                                         seda_rule=code)
            rule = cnx.create_entity('SEDAAccessRule',
                                     seda_access_rule=transfer,
                                     seda_seq_access_rule_rule=rule_seq)
            cnx.commit()
            self.assertMDNow(transfer)
            # should not have touched scheme's date
            self.assertEqual(scheme.modification_date, YESTERDAY)
            self.resetMD(cnx, transfer)
            # edit link from composite children to an entity which is not part of the container
            rule_seq.cw_set(seda_rule=None)
            cnx.commit()
            self.assertMDNow(transfer)
            self.resetMD(cnx, transfer)
            # deletion of a composite children
            rule.cw_delete()
            cnx.commit()
            self.assertMDNow(transfer)
            transfer.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            self.resetMD(cnx, transfer)
            # edit link from root to an entity which is not part of the container
            organization_unit = testutils.organization_unit(
                cnx, u'marcel', archival_roles=['deposit'])
            transfer.cw_set(reverse_use_profile=organization_unit)
            cnx.commit()
            self.assertEqual(transfer.modification_date, YESTERDAY)

    def test_externaluri_to_concept(self):
        with self.admin_access.repo_cnx() as cnx:
            # create some authority record and related objects
            agent = testutils.authority_record(cnx, u'bob')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            place = cnx.create_entity('AgentPlace', place_address=address, place_agent=agent)
            function = cnx.create_entity('AgentFunction', name=u'sponge', function_agent=agent)
            info = cnx.create_entity('LegalStatus', legal_status_agent=agent)
            cnx.commit()
            # create some external uri and link it to place, function and information entities
            exturi = cnx.create_entity('ExternalUri', cwuri=u'http://someuri/someobject',
                                       uri=u'http://someuri/someobject',
                                       reverse_equivalent_concept=[place, function, info])
            cnx.commit()
            # now insert a concept with the external uri as cwuri
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object', cwuri=u'http://someuri/someobject')
            cnx.commit()
            # ensure the external uri has been replaced by the concept and deleted
            place.cw_clear_all_caches()
            self.assertEqual(place.equivalent_concept[0].eid, concept.eid)
            function.cw_clear_all_caches()
            self.assertEqual(function.equivalent_concept[0].eid, concept.eid)
            info.cw_clear_all_caches()
            self.assertEqual(info.equivalent_concept[0].eid, concept.eid)
            self.assertFalse(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_externaluri_to_authorityrecord_subject(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            exturi = cnx.create_entity('ExternalUri', cwuri=u'a/b/c', uri=u'a/b/c')
            arelation = cnx.create_entity('AssociationRelation',
                                          association_from=bob,
                                          association_to=exturi)
            cnx.commit()
            alice = testutils.authority_record(cnx, u'alice', cwuri=u'a/b/c')
            cnx.commit()
            arelation.cw_clear_all_caches()
            self.assertEqual(arelation.association_to[0], alice)
            self.assertFalse(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_externaluri_to_authorityrecord_object(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            exturi = cnx.create_entity('ExternalUri', cwuri=u'a/b/c', uri=u'a/b/c')
            arelation = cnx.create_entity('AssociationRelation',
                                          association_from=exturi,
                                          association_to=bob)
            cnx.commit()
            alice = testutils.authority_record(cnx, u'alice', cwuri=u'a/b/c')
            cnx.commit()
            arelation.cw_clear_all_caches()
            self.assertEqual(arelation.association_from[0], alice)
            self.assertFalse(cnx.execute('Any X WHERE X eid %(x)s', {'x': exturi.eid}))

    def test_user_email(self):
        # EmailAddress belong to the Organisation graph, hence may conflict with hooks when linked
        # to a user. This is a non-regression test checking that no error is raised on attempt to
        # add an email to a user (which is not adaptable to IContained nor IContainer)
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('EmailAddress', address=u'admin@cubicweb.org',
                              reverse_use_email=cnx.user.eid)
            cnx.commit()

    def test_sortable_index_is_eid_on_creation(self):
        with self.admin_access.cnx() as cnx:
            arecord = testutils.authority_record(cnx, u'test')
            mandate = cnx.create_entity('Mandate', term=u'casting', mandate_agent=arecord)
            cnx.commit()
            self.assertEqual(mandate.eid, mandate.index)


class ARKGenerationHooksTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def _check_ark(self, entity, parent_ark=None):
        self.assertTrue(testutils.match_ark(entity.ark), entity.ark)
        if parent_ark is not None:
            self.assertTrue(entity.ark.startswith(parent_ark + '/'),
                            (parent_ark, entity.ark))

    def _check_ark_and_cwuri(self, entity, **kwargs):
        self._check_ark(entity, **kwargs)
        self.assertTrue(testutils.match_ark_uri(entity.cwuri), entity.cwuri)

    def test_external_ark_invalid(self):
        with self.admin_access.cnx() as cnx:
            for ark in [
                u'string/name',  # NAAN is not a number
                u'123',  # Missing Name
            ]:
                with self.subTest(ark=ark):
                    with self.assertRaises(ValidationError) as cm:
                        testutils.authority_record(cnx, u'bob', ark=ark)
                    self.assertIn('malformatted ARK idenfitier', str(cm.exception))

    def test_external_ark_duplicated(self):
        ark = '1/bob/thefirst'
        with self.admin_access.cnx() as cnx:
            cnx.create_entity('Agent', name=u'bob', ark=ark)
            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity('ConceptScheme', ark=ark)
            self.assertIn('already exists', str(cm.exception))

    def test_external_ark_duplicated_generated(self):
        with self.admin_access.cnx() as cnx:
            authority = testutils.authority_with_naa(cnx)
            agent = cnx.create_entity('Agent', name=u'bob', authority=authority)
            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity('ConceptScheme', ark=agent.ark)
            self.assertIn('already exists', str(cm.exception))

    def test_ark_generation_authorityrecord(self):
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'bob')
            cnx.commit()
            self._check_ark_and_cwuri(agent)
            agent = testutils.authority_record(cnx, u'john', ark=u'999/123456')
            cnx.commit()
            self.assertEqual(agent.ark, '999/123456')
            self.assertEqual(agent.cwuri, 'ark:/999/123456')
            agent = testutils.authority_record(cnx, u'alf', cwuri=u'http://someuri/someagent')
            cnx.commit()
            self._check_ark(agent)
            self.assertEqual(agent.cwuri, 'http://someuri/someagent')

    def test_ark_generation_seda_profile(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_profile(cnx)
            self._check_ark_and_cwuri(profile)

    def test_ark_generation_seda_profile_ark_given(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_profile(cnx, ark=u'99/124')
            self.assertEqual(profile.ark, '99/124')
            self.assertEqual(profile.cwuri, 'ark:/99/124')

    def test_ark_generation_seda_profile_with_cwuri(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_profile(
                cnx, cwuri=u'http://example.org/profile/125')
            self._check_ark(profile)
            self.assertEqual(profile.cwuri, 'http://example.org/profile/125')

    def test_ark_generation_seda_archiveunit(self):
        with self.admin_access.repo_cnx() as cnx:
            naa = testutils.naa(cnx)
            unit, _, _ = testutils.create_archive_unit(None, cnx=cnx, ark_naa=naa)
            self._check_ark_and_cwuri(unit)

    def test_ark_generation_seda_archiveunit_in_archivetransfer(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_profile(cnx)
            unit, _, _ = testutils.create_archive_unit(profile, cnx)
            self._check_ark_and_cwuri(unit)
            self.assertTrue(unit.ark.startswith('0/'), unit.ark)

    def test_ark_generation_concept(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object')
            cnx.commit()
            self._check_ark_and_cwuri(scheme)
            self._check_ark_and_cwuri(concept, parent_ark=scheme.ark)
            scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://someuri/somescheme',
                                       ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object', cwuri=u'http://someuri/someconcept')
            cnx.commit()
            self._check_ark(scheme)
            self.assertEqual(scheme.cwuri, 'http://someuri/somescheme')
            self._check_ark(concept, scheme.ark)
            self.assertEqual(concept.cwuri, 'http://someuri/someconcept')
            scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://dcf/res/ark:/67717/Matiere',
                                       ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'some object', cwuri=u'http://dcf/res/ark:/67717/1234')
            cnx.commit()
            self.assertEqual(scheme.ark, '67717/Matiere')
            self.assertEqual(scheme.cwuri, 'http://dcf/res/ark:/67717/Matiere')
            self.assertEqual(concept.ark, '67717/1234')
            self.assertEqual(concept.cwuri, 'http://dcf/res/ark:/67717/1234')

    def test_ark_generation_organization_unit(self):
        with self.admin_access.cnx() as cnx:
            authority = testutils.authority_with_naa(cnx)
            orgunit = testutils.organization_unit(
                cnx, u'test', authority=authority)
            cnx.commit()
            self._check_ark(orgunit, authority.ark)


class AuthorityRecordHookTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            self.org_eid = testutils.authority_with_naa(cnx).eid
            self.orgunit_eid = testutils.organization_unit(
                cnx, u'test', archival_roles=('archival', ),
                reverse_archival_unit=self.org_eid, authority=self.org_eid).eid
            self.create_user(cnx, u'bob', authority=self.org_eid)
            cnx.commit()

    def test_automatic_use_authorityrecord(self):
        with self.new_access(u'bob').cnx() as cnx:
            arecord = testutils.authority_record(cnx, u'test')
            cnx.commit()
            self.assertEqual(
                [ou.eid for ou in arecord.reverse_use_authorityrecord],
                [self.orgunit_eid])

    def test_automatic_use_authorityrecord_no_archival_unit(self):
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, u'alice')
            cnx.commit()
        with self.new_access(u'alice').cnx() as cnx:
            arecord = testutils.authority_record(cnx, u'test')
            cnx.commit()
            arecord.cw_clear_all_caches()
            self.assertFalse(arecord.reverse_use_authorityrecord)


class ExtractArkTC(unittest.TestCase):

    configcls = PostgresApptestConfiguration

    def test_ok(self):
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234'), '67717/1234')
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234#something'), '67717/1234')
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234?value'), '67717/1234')
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717/1234/sub'), '67717/1234')
        self.assertEqual(extract_ark('ark:/67717/1234'), '67717/1234')

    def test_ko(self):
        self.assertEqual(extract_ark('http://dcf/res/ark:/67717'), None)
        self.assertEqual(extract_ark('http://someuri/67717/1234'), None)


class EntityLifeCycleTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            bob = self.create_user(cnx, u'bob',
                                   authority=testutils.authority_with_naa(cnx))
            cnx.commit()
            self.bob_eid = bob.eid

    def _check_create(self, cnx, eid, msg):
        activity = cnx.find('Activity', type=u'create', used=eid).one()
        self.assertEqual(activity.associated_with[0].eid, cnx.user.eid)
        self.assertEqual(activity.generated[0].eid, eid)
        self.assertEqual(activity.description, msg)

    def _check_modification(self, cnx, eid, msg):
        activity = cnx.find('Activity', type=u'modify', used=eid).one()
        self.assertEqual(activity.associated_with[0].eid, self.bob_eid)
        self.assertEqual(activity.used[0].eid, eid)
        self.assertEqual(activity.generated[0].eid, eid)
        self.assertEqual(activity.description, msg)

    def test_authorityrecord_create_update(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            cnx.commit()
            self._check_create(cnx, authorityrecord.eid,
                               'created authorityrecord')
            authorityrecord.cw_set(isni=u'123', record_id=u'Adam')
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'modified isni, record_id')

    def test_authorityrecord_add_component(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            cnx.commit()
            cnx.create_entity('AgentFunction', name=u'secret agent',
                              function_agent=authorityrecord)
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'added function_agent_object')

    def test_authorityrecord_delete_component(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            af = cnx.create_entity('AgentFunction', name=u'secret',
                                   function_agent=authorityrecord)
            cnx.commit()
            af.cw_delete()
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'removed function_agent_object')

    def test_authorityrecord_update_component(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            pn = cnx.create_entity('AgentFunction', name=u'secret agent',
                                   function_agent=authorityrecord)
            cnx.commit()
            pn.cw_set(name=u'boss of secret agents')
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'modified function_agent_object')

    def test_authorityrecord_add_subcomponent(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            cnx.commit()
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'added place_address, place_agent_object')

    def test_authorityrecord_delete_subcomponent(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            cnx.commit()
            address.cw_delete()
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid, 'removed place_address')

    def test_authorityrecord_update_subcomponent(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            cnx.commit()
            address.cw_set(street=u"1 avenue de l'Europe")
            cnx.commit()
            self._check_modification(cnx, authorityrecord.eid,
                                     'modified place_address')

    def test_authorityrecord_multi_modification(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            authorityrecord = testutils.authority_record(cnx, u'Smith')
            f1 = cnx.create_entity('AgentFunction', name=u'special agent',
                                   function_agent=authorityrecord)
            gc = cnx.create_entity('GeneralContext', content=u'sunglasses',
                                   general_context_of=authorityrecord)
            cnx.commit()
            gc.cw_set(content=u'stetson')
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', place_address=address,
                              place_agent=authorityrecord)
            f1.cw_delete()
            cnx.create_entity('AgentFunction', name=u'mib',
                              function_agent=authorityrecord)
            authorityrecord.cw_set(isni=u'123', record_id=u'Adam')
            cnx.commit()
            self._check_modification(
                cnx, authorityrecord.eid,
                '* modified function_agent_object, general_context_of_object, isni, record_id\n'
                '* added place_address, place_agent_object')

    def test_authorityrecord2authorityrecord_relation(self):
        with self.new_access(u'bob').repo_cnx() as cnx:
            k = testutils.authority_record(cnx, u'authorityrecord K')
            d = testutils.authority_record(cnx, u'authorityrecord D')
            cnx.commit()
            cnx.create_entity('AssociationRelation', association_from=k, association_to=d)
            cnx.commit()
            self._check_modification(cnx, k.eid, 'added association_from_object')
            self._check_modification(cnx, d.eid, 'added association_to_object')

    def test_no_activity_generated(self):
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.agent(cnx, u'smith', authority=cnx.find('Organization')[0][0])
            cnx.commit()
            authorityrecord = testutils.authority_record(
                cnx, u'Smith', reverse_authority_record=agent)
            cnx.commit()
            agent.cw_set(name=u'jones')
            cnx.commit()
            self.assertFalse(cnx.find('Activity', type=u'modify',
                                      used=authorityrecord.eid))

    def test_concept(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus',
                                       ark_naa=testutils.naa(cnx))
            cnx.commit()
            # concept scheme creation
            self._check_create(cnx, scheme.eid, 'created conceptscheme')

        with self.new_access(u'bob').repo_cnx() as cnx:
            scheme = cnx.entity_from_eid(scheme.eid)

            concept = scheme.add_concept(u'hello')
            cnx.commit()
            # concept scheme add concept
            self._check_modification(cnx, scheme.eid, 'added in_scheme_object')
            # concept creation
            self._check_create(cnx, concept.eid, 'created concept')
            subconcept = concept.add_concept(u'goodbye')
            cnx.commit()
            # concept add concept
            self._check_modification(cnx, concept.eid, 'added broader_concept_object')
            # subconcept creation
            self._check_create(cnx, subconcept.eid, 'created concept')
            # subconcept creation scheme
            rset = cnx.execute('Activity X ORDERBY X DESC WHERE X type "modify", X used %(x)s',
                               {'x': scheme.eid})
            self.assertEqual(len(rset), 2)
            self.assertEqual(rset.get_entity(0, 0).description, 'added in_scheme_object')

    def test_archive_transfer_create_update(self):
        with self.new_access(u'bob').cnx() as cnx:
            transfer = testutils.setup_profile(cnx)
            cnx.commit()
            self._check_create(cnx, transfer.eid, 'created sedaarchivetransfer')
            transfer.cw_set(title=u'123')
            cnx.commit()
            self._check_modification(cnx, transfer.eid,
                                     'modified title')

    def test_nonregr_concept_container(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'my thesaurus',
                                       ark_naa=testutils.naa(cnx))
            concept = scheme.add_concept(u'hello')
            cnx.commit()
            concept.cw_delete()
            cnx.commit()

    def test_inter_containers(self):
        with self.admin_access.repo_cnx() as cnx:
            scheme = testutils.scheme_for_type(
                cnx, 'seda_rule', 'SEDASeqAccessRuleRule', u'AR038')
            code = scheme.reverse_in_scheme[0]
            transfer = testutils.setup_profile(cnx)
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            cnx.commit()

            nb_scheme_activities = len(cnx.find('Activity', used=scheme.eid))
            nb_concept_activities = len(cnx.find('Activity', used=code.eid))
            nb_transfer_activities = len(cnx.find('Activity', used=transfer.eid))

            rule_seq = cnx.create_entity(
                'SEDASeqAccessRuleRule',
                reverse_seda_start_date=cnx.create_entity('SEDAStartDate'),
                seda_rule=code)
            cnx.create_entity(
                'SEDAAccessRule',
                seda_access_rule=transfer,
                seda_seq_access_rule_rule=rule_seq)
            cnx.commit()

            self.assertEqual(len(cnx.find('Activity', used=code.eid)),
                             nb_concept_activities)
            self.assertEqual(len(cnx.find('Activity', used=transfer.eid)),
                             nb_transfer_activities + 1)

            cnx.create_entity('SEDAMimeTypeCodeListVersion',
                              seda_mime_type_code_list_version_from=transfer,
                              seda_mime_type_code_list_version_to=scheme)
            cnx.commit()

            self.assertEqual(len(cnx.find('Activity', used=scheme.eid)),
                             nb_scheme_activities)
            self.assertEqual(len(cnx.find('Activity', used=transfer.eid)),
                             nb_transfer_activities + 2)


class SEDAArchiveTransferHooksTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_profile_deprecated_by_published_successor(self):
        """Test hook deprecating a SEDA Profile upon successor publication."""
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.setup_profile(cnx)
            cnx.commit()
            workflow = profile.cw_adapt_to('IWorkflowable')
            workflow.fire_transition('publish')
            cnx.commit()
            profile.cw_clear_all_caches()
            self.assertEqual(workflow.state, 'published')
            cloned = cnx.create_entity('SEDAArchiveTransfer', title=u'Clone',
                                       new_version_of=profile,
                                       ark_naa=testutils.naa(cnx))
            cnx.commit()
            cloned.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            profile.cw_clear_all_caches()
            self.assertEqual(workflow.state, 'deprecated')

    def test_profile_deprecated_unlinked_from_organization_unit(self):
        with self.admin_access.cnx() as cnx:
            profile = testutils.setup_profile(cnx)
            organization_unit = testutils.organization_unit(
                cnx, u'org', archival_roles=[u'deposit'])
            cnx.commit()
            workflow = profile.cw_adapt_to('IWorkflowable')
            workflow.fire_transition('publish')
            cnx.commit()
            organization_unit.cw_set(use_profile=profile)
            cnx.commit()
            workflow.fire_transition('deprecate')
            cnx.commit()
            self.assertFalse(organization_unit.use_profile)


class ConceptSchemeWorkflowPermissionsTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_cannot_deprecate_used_vocabulary(self):
        with self.admin_access.cnx() as cnx:
            scheme = testutils.scheme_for_type(cnx, u'seda_keyword_type_to', None,
                                               u'type1')
            cnx.commit()
            scheme.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            assert scheme.scheme_relation_type
            with self.assertRaises(ValidationError) as cm:
                scheme.cw_adapt_to('IWorkflowable').fire_transition('deprecate')
            cnx.rollback()
            self.assertIn(
                'this concept scheme is used in at least one SEDA constraint',
                str(cm.exception))


if __name__ == '__main__':
    unittest.main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref custom widgets"""

import calendar
from datetime import datetime

from six import text_type
from dateutil.parser import parse as parse_date

from logilab.common.date import ustrftime

from cubicweb import tags
from cubicweb.utils import JSString, json_dumps
from cubicweb.uilib import js
from cubicweb.web import ProcessFormError, formwidgets as fw
from cubicweb.web.views import ajaxcontroller


def process_incomplete_date(datestr, end=False):
    """Parse a date from a string.

    Return the date if the given string is a valid date, else if only the year or the year and the
    month is given, return a date completed with default day and/or month value:

    * if end is True, default day/month is 31/12 and default day is the last day of the given month
    * if end is False, default day/month is 01/01 and default day is 1
    """
    if not datestr:
        return None
    if end:
        if len(datestr.split("/")) == 2:  # = day is not defined
            pdate = parse_date(datestr, default=datetime(9999, 12, 1), dayfirst=True)
            last_day = calendar.monthrange(pdate.year, pdate.month)[1]
            pdate = pdate.replace(day=last_day)
        else:
            pdate = parse_date(datestr, default=datetime(9999, 12, 31), dayfirst=True)
    else:
        pdate = parse_date(datestr, default=datetime(9999, 1, 1), dayfirst=True)
    if pdate.year == 9999:
        raise ValueError()
    return pdate.date()


class JQueryIncompleteDatePicker(fw.JQueryDatePicker):
    """Custom datefield enabling the user to enter only the year or the month and year
    of the date.

    If only the year or the year and the month is given, behaviour will change depending on the
    `default_end` boolean flag:

    * when true, default month is december and default day is the last day of the given month
    * when false, default month is january and default day is 1
    """
    def __init__(self, *args, **kwargs):
        self.default_end = kwargs.pop("default_end", False)
        self.update_min = kwargs.pop("update_min", None)
        self.update_max = kwargs.pop("update_max", None)
        super(JQueryIncompleteDatePicker, self).__init__(*args, **kwargs)

    def process_field_data(self, form, field):
        datestr = form._cw.form.get(field.input_name(form))
        try:
            return process_incomplete_date(datestr, self.default_end)
        except ValueError as exc:
            if "day is out of range for month" in exc:
                raise ProcessFormError(form._cw._("day is out of range for month"))
            elif "month must be in 1..12" in exc:
                raise ProcessFormError(form._cw._("month must be in 1..12"))
            else:
                raise ProcessFormError(form._cw._("can not process %s: expected format "
                                                  "dd/mm/yyyy, mm/yyyy or yyyy" % datestr))

    def attributes(self, form, field):
        attrs = super(JQueryIncompleteDatePicker, self).attributes(form, field)
        form._cw.add_js('cubes.saem_ref.js')
        if self.update_max:
            domid = '%s-subject:%s' % (self.update_max, form.edited_entity.eid)
            attrs['onChange'] = js.saem.update_datepicker_minmax(domid, 'maxDate',
                                                                 JSString('this.value'))
        if self.update_min:
            domid = '%s-subject:%s' % (self.update_min, form.edited_entity.eid)
            attrs['onChange'] = js.saem.update_datepicker_minmax(domid, 'minDate',
                                                                 JSString('this.value'))
        return attrs

    def _render(self, form, field, renderer):
        req = form._cw
        if req.lang != 'en':
            req.add_js('jquery.ui.datepicker-%s.js' % req.lang)
        domid = field.dom_id(form, self.suffix)
        # XXX find a way to understand every format
        fmt = req.property_value('ui.date-format')
        picker_fmt = fmt.replace('%Y', 'yy').replace('%m', 'mm').replace('%d', 'dd')
        max_date = min_date = None
        if self.update_min:
            current = getattr(form.edited_entity, self.update_min)
            if current is not None:
                max_date = ustrftime(current, fmt)
        if self.update_max:
            current = getattr(form.edited_entity, self.update_max)
            if current is not None:
                min_date = ustrftime(current, fmt)
        req.add_onload(u'cw.jqNode("%s").datepicker('
                       '{buttonImage: "%s", dateFormat: "%s", firstDay: 1,'
                       ' showOn: "button", buttonImageOnly: true, minDate: %s, maxDate: %s});'
                       % (domid, req.uiprops['CALENDAR_ICON'], picker_fmt, json_dumps(min_date),
                          json_dumps(max_date)))
        return self._render_input(form, field)


class ConceptAutoCompleteWidget(fw.TextInput):
    """Derive from simple text input to create an autocompletion widget if some scheme is specified.
    Otherwise, free text is fine if `optional` argument is true. In such case:

    * `slave_name` is expected to be the name of the text attribute

    * you'll have to use this widget from a custom field that will handle the relation to the
      concept(see example in agent views).
    """

    def __init__(self, slave_name, master_name, optional=False, **kwargs):
        super(ConceptAutoCompleteWidget, self).__init__(**kwargs)
        self.slave_name = slave_name
        self.master_name = master_name
        self.optional = optional

    def _render(self, form, field, render):
        entity = form.edited_entity
        slave_id = field.dom_id(form, self.suffix)
        master_id = slave_id.replace(self.slave_name, self.master_name)
        if entity.has_eid():
            concept = entity.concept
        else:
            concept = None
        req = form._cw
        req.add_js(('cubicweb.js', 'cubicweb.ajax.js', 'cubes.saem_ref.js'))
        req.add_onload(js.saem.initSchemeConceptFormField(master_id, slave_id))
        if concept is None:
            value = getattr(entity, self.slave_name) if self.optional else None
            eid = u''
        else:
            if concept.cw_etype == 'Concept':
                value = concept.label()
            else:  # ExternalURI
                value = concept.dc_title()
            eid = text_type(concept.eid)
        # we need an hidden input to handle the value while the text input display the label
        inputs = [
            tags.input(name=field.input_name(form, 'Label'), id=slave_id + 'Label',
                       klass='form-control', type='text',
                       value=value),
            tags.input(name=field.input_name(form), id=slave_id, type='hidden',
                       value=eid)
        ]
        return u'\n'.join(inputs)


@ajaxcontroller.ajaxfunc(output_type='json')
def keyword_concepts(self):
    assert self._cw.form['scheme']
    scheme = int(self._cw.form['scheme'])
    term = self._cw.form['q']
    limit = self._cw.form.get('limit', 50)
    return [{'value': eid, 'label': label}
            for eid, label in self._cw.execute(
                'DISTINCT Any C,N ORDERBY N LIMIT %s WHERE C in_scheme S, S eid %%(s)s, '
                'C preferred_label L, L label N, L label ILIKE %%(term)s' % limit,
                {'s': scheme, 'term': u'%%%s%%' % term})]



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views/forms/actions/components for web ui"""

from functools import wraps

from six import text_type

from logilab.common.decorators import monkeypatch
from logilab.mtconverter import xml_escape
from logilab.common.registry import yes

from cubicweb import _
from cubicweb import NoSelectableObject, neg_role, tags
from cubicweb.utils import json_dumps, js_href
from cubicweb.uilib import domid, js
from cubicweb.view import EntityView
from cubicweb.web.views import ajaxcontroller
from cubicweb.predicates import (has_permission, is_instance, multi_lines_rset, match_kwargs,
                                 partial_has_related_entities)
from cubicweb.web import component, formwidgets as fw
from cubicweb.web.views import (
    basecomponents,
    baseviews,
    primary,
    tabs,
    uicfg,
    urlpublishing,
    urlrewrite,
)

from cubicweb_squareui.views.basetemplates import basetemplates
from cubicweb_seda.views import dropdown_button, has_rel_perm, widgets as sedawidgets

from .. import cwuri_url


pvs = uicfg.primaryview_section
afs = uicfg.autoform_section
abaa = uicfg.actionbox_appearsin_addmenu
affk = uicfg.autoform_field_kwargs

pvs.tag_subject_of(('*', 'ark_naa', '*'), 'attributes')
afs.tag_subject_of(('*', 'ark_naa', '*'), 'main', 'attributes')
afs.tag_subject_of(('*', 'custom_workflow', '*'), 'main', 'hidden')
afs.tag_subject_of(('*', 'primary_email', '*'), 'main', 'hidden')

abaa.tag_object_of(('*', 'for_user', '*'), False)  # do not expose CWProperties mecanism
abaa.tag_subject_of(('*', 'use_email', '*'), False)
abaa.tag_subject_of(('*', 'phone_number', '*'), False)

affk.set_field_kwargs('*', 'ark', widget=fw.TextInput({'size': 80}), required=False)


def external_link(text, href):
    """Return an HTML link with an icon indicating that the URL is external.
    """
    icon = tags.span(klass="glyphicon glyphicon-share-alt")
    return tags.a(u' '.join([text, icon]), href=href, escapecontent=False)


def add_etype_link(req, etype, text=u'', klass='icon-plus-circled pull-right',
                   **urlparams):
    """Return an HTML link to add an entity of type 'etype'."""
    vreg = req.vreg
    eschema = vreg.schema.eschema(etype)
    if eschema.has_perm(req, 'add'):
        url = vreg['etypes'].etype_class(etype).cw_create_url(req, **urlparams)
        return tags.a(text, href=url, klass=klass,
                      title=req.__('New %s' % etype))
    return u''


def import_etype_link(req, etype, url):
    """Return an HTML link to the view that may be used to import an entity of type `etype`.
    """
    eschema = req.vreg.schema.eschema(etype)
    if eschema.has_perm(req, 'add'):
        return tags.a(u'', href=url, klass='icon-upload pull-right',
                      title=req.__('Import %s' % etype))
    return u''


def add_relations_button(req, entity, msg, *relations_info, **extraurlparams):
    """Return an HTML dropdown button to add relations with `entity` as object"""
    links = [relation_link(req, entity, relation_info, **extraurlparams)
             for relation_info in relations_info]
    links = [l for l in links if l]
    if links:
        # No links if user cannot add any relation.
        return dropdown_button(req._(msg), links)


def relation_link(req, entity, relation_info, **extraurlparams):
    """Return a HTML link to add a `rtype` relation to `entity`"""
    rtype, role, target_etype, label = relation_info
    rschema = req.vreg.schema[rtype]
    targets = rschema.targets(role=role)
    if target_etype is not None:
        for target in targets:
            etype = target.type
            if etype == target_etype:
                break
        else:
            raise AssertionError('{0} not found in targets of {1}'.format(
                target_etype, rschema))
    else:
        assert len(targets) == 1, \
            'expecting a single {0} for relation {1}'.format(role, rschema)
        etype = targets[0].type
    if not has_rel_perm('add', entity, rtype, role, target_etype=etype):
        return u''
    linkto = '{rtype}:{eid}:{role}'.format(rtype=rtype, eid=entity.eid, role=neg_role(role))
    urlparams = {'__linkto': linkto,
                 '__redirectpath': entity.rest_path()}
    urlparams.update(extraurlparams)
    if label is None:
        label = etype
    return add_etype_link(req, etype, text=req._(label).lower(),
                          klass='', **urlparams)


def editlinks(icon_info):
    """Decorator for `entity_call` to add "edit" links."""
    def decorator(entity_call):
        @wraps(entity_call)
        def wrapper(self, entity, **kwargs):
            editurlparams = {
                '__redirectpath': kwargs.pop('__redirectpath',
                                             entity.rest_path()),
            }
            if 'tabid' in kwargs:
                editurlparams['__redirectparams'] = 'tab=' + kwargs.pop('tabid')
            html_icons = entity.view('saem.edit-link',
                                     icon_info=icon_info,
                                     editurlparams=editurlparams)
            if html_icons:
                self.w(tags.div(html_icons, klass='pull-right'))
            return entity_call(self, entity, **kwargs)
        return wrapper
    return decorator


class ExternalLinkView(EntityView):
    """Render an entity as an external link according to `rtype` keyword
    argument or automatically if it is an ExternalUri.
    """

    __regid__ = 'saem.external_link'
    __select__ = EntityView.__select__ & (
        is_instance('ExternalUri') | match_kwargs('rtype'))

    def entity_call(self, entity, rtype=None):
        if rtype is None:
            rtype = 'uri'  # ExternalUri
        href = text = getattr(entity, rtype)
        self.w(external_link(text, href))


def configure_relation_widget(req, div, search_url, title, multiple, validate):
    """Build a javascript link to invoke a relation widget

    Widget will be linked to div `div`, with a title `title`. It will display selectable entities
    matching url `search_url`. bool `multiple` indicates whether several entities can be selected or
    just one, `validate` identifies the javascript callback that must be used to validate the
    selection.
    """
    req.add_js('cubes.saem_ref.js')
    return sedawidgets.configure_relation_widget(req, div, search_url, title, multiple, validate)


ENTITY_CREATION_KWARGS = {
    'SEDAArchiveUnit': {'unit_type': 'unit_content'},
}


class AddEntityComponent(component.CtxComponent):
    """Component with 'add' link to be displayed in 'same etype' views usually 'SameETypeListView'.
    """
    __regid__ = 'saem.add_entity'
    __select__ = (component.CtxComponent.__select__ & multi_lines_rset() & has_permission('add')
                  & is_instance('AuthorityRecord', 'ConceptScheme',
                                'SEDAArchiveTransfer', 'SEDAArchiveUnit'))
    context = 'navtop'

    def render_body(self, w):
        etype = self.cw_rset.description[0][0]
        w(add_etype_link(self._cw, etype, **ENTITY_CREATION_KWARGS.get(etype, {})))


class ImportEntityComponent(component.CtxComponent):
    """Component with 'import' link to be displayed in 'same etype' views usually
    'SameETypeListView'.

    Concret class should fill the `import_vid` class attribute and add a proper `is_instance`
    selector.
    """
    __abstract__ = True
    __regid__ = 'saem.import_entity'
    __select__ = component.CtxComponent.__select__ & multi_lines_rset() & has_permission('add')
    import_url = None  # URL of the view that may be used to import data
    context = 'navtop'

    def render_body(self, w):
        etype = self.cw_rset.description[0][0]
        w(import_etype_link(self._cw, etype, self.import_url))


# Bootstrap configuration.
basetemplates.TheMainTemplate.twbs_container_cls = 'container'


# Wrap breadcrumbs items within a "container" div.
@monkeypatch(basetemplates.HTMLPageHeader)
def breadcrumbs(self, view):
    components = self.get_components(view, context='header-center')
    if components:
        self.w(u'<nav role="navigation" class="cw-breadcrumb">')
        for comp in components:
            self.w(u'<div class="%s">' % basetemplates.TheMainTemplate.twbs_container_cls)
            comp.render(w=self.w)
            self.w(u'</div>')
        self.w(u'</nav>')


# Display header-right components in a specific div, not using 'nav' bootstrap style and handle an
# 'active' class in the nav component (main-navigation)
@monkeypatch(basetemplates.HTMLPageHeader)
def main_header(self, view):
    w = self.w
    w(u'<nav class="navbar navbar-default" role="banner">')
    w(u'<div class="container">')
    w(u'<div class="pull-right dropdown">')
    components = self.get_components(view, context='header-right')
    for comp in components:
        comp.render(w=w)
    w(u'</div>')
    self.display_navbar_header(w, view)
    w(u'<div id="tools-group" class="collapse navbar-collapse">')
    components = self.get_components(view, context='main-navigation')
    if components:
        w(u'<ul class="nav navbar-nav navbar-left">')
        for comp in components:
            if getattr(comp, 'active', False):
                w(u'<li class="active">')
            else:
                w(u'<li>')
            comp.render(w=w)
            w(u'</li>')
        w(u'</ul>')
    w(u'</div></div></nav>')


class MainNavigationMenuComponent(basecomponents.HeaderComponent):
    __abstract__ = True
    __select__ = yes()
    context = 'main-navigation'
    path = label = None

    @property
    def active(self):
        return self._cw.relative_path().lower() == self.path

    def render(self, w, **kwargs):
        self._cw.add_css('cubes.saem_ref.css')
        w(tags.a(self._cw._(self.label), href=self._cw.build_url(self.path)))


def main_navigation_menu():
    """Yield navigation component instance to fill the navigation menu"""
    for order, (path, label) in enumerate([
        ('Organization', _('Organization_plural_short')),
        ('AuthorityRecord', 'AuthorityRecord_plural'),
        ('ConceptScheme', _('Vocabularies')),
        ('SEDAArchiveTransfer', 'SEDAArchiveTransfer_plural'),
        ('sedalib', _('SEDA components')),
    ]):
        yield type(path + 'NavMenuEntryComponent', (MainNavigationMenuComponent,), {
            '__regid__': 'saem.navmenu_' + path.lower(),
            'path': path.lower(),
            'label': label,
            'order': order,
        })


class SAEMHTMLPageFooter(basetemplates.HTMLPageFooter):
    def footer_content(self):
        self.w(u'''<p>
  <a href="http://www.gironde.fr" target="_blank"><img src="{logo_cd33}" /></a>
  <a href="http://www.bordeaux-metropole.fr" target="_blank"><img src="{logo_metro}" /></a>
  <a href="http://www.bordeaux.fr" target="_blank"><img src="{logo_mairie}" /></a>
</p>'''.format(logo_cd33=self._cw.data_url('cd33.jpg'),
               logo_metro=self._cw.data_url('bx_metro.jpg'),
               logo_mairie=self._cw.data_url('mairie_bx.jpg')))


class SortableListView(baseviews.ListView):
    __regid__ = 'sortable-list'
    __select__ = baseviews.ListView.__select__ & match_kwargs('listid')

    def call(self, **kwargs):
        self._cw.add_js('cubes.saem_ref.js')
        listid = kwargs['listid']
        self._cw.add_onload(
            js.saem.makeSortable(listid, kwargs['rtype'])
        )
        super(SortableListView, self).call(**kwargs)

    def cell_call(self, row, col=0, vid=None, klass=None, **kwargs):
        self.w(u'<li data-eid={} >'.format(self.cw_rset[row][0]))
        self.wview(self.item_vid, self.cw_rset, row=row, col=col, vid=vid, **kwargs)
        self.w(u'</li>')


class RelatedEntitiesListView(EntityView):
    """Abstract entity view for displaying an related entities in a list.

    This view should be called with a ``subvid`` parameter indicating the ``regid`` of the view
    to be used for each related entity.
    """

    __abstract__ = True
    __select__ = EntityView.__select__ & partial_has_related_entities()
    rtype = None
    role = 'object'
    subvid = 'saem.listitem'
    listvid = 'list'
    target_etype = None
    subvid_kwargs = None

    def related_rset(self, entity):
        if self.target_etype is not None:
            targettypes = (self.target_etype, )
        else:
            targettypes = None
        return entity.related(self.rtype, role=self.role,
                              targettypes=targettypes)

    @property
    def title(self):
        return self.rtype + '_object' if self.role == 'object' else self.rtype

    def entity_call(self, entity, **kwargs):
        kwargs.update(self.subvid_kwargs or {})
        kwargs['__redirectpath'] = entity.rest_path()
        if self.title:
            self.w(tags.h2(self._cw._(self.title)))
        rset = self.related_rset(entity)
        if len(rset) == 1:
            self._cw.view(self.subvid, rset=rset, w=self.w, **kwargs)
        else:
            # assign divid to self for reuse in page_navigation_url below
            self.divid = u'{}{}'.format(domid(self.__regid__), entity.eid)
            self.w(u'<div id="{}">'.format(self.divid))
            self.do_paginate(rset=rset)
            listid = self.divid + "_list"
            kwargs['rtype'] = self.rtype
            self._cw.view(self.listvid, rset=rset, w=self.w,
                          subvid=self.subvid, listid=listid, **kwargs)
            self.w(u'</div>')

    def page_navigation_url(self, navcomp, _path, params):
        divid = params['divid'] = self.divid
        # we don't want vid from _cw.form which target the parent tab
        params['vid'] = self.__regid__
        # generate rql instead of using cw_rset.printable_rql since it's usually an ambiguous query
        # using ARK identifier
        params['rql'] = 'Any X WHERE X eid {}'.format(self.cw_rset.one().eid)
        return js_href("$(%s).loadxhtml(AJAX_PREFIX_URL, %s, 'get', 'swap')" % (
            json_dumps('#' + divid), js.ajaxFuncArgs('view', params)))


class RelationInfo(tuple):

    def __new__(cls, rtype, role, targets=None, label=None):
        return tuple.__new__(cls, (rtype, role, targets, label))


class SubviewsTabView(tabs.TabsMixin, EntityView):
    """Abstract tab view rendering views listed in `subvids` attribute."""
    __abstract__ = True
    subvids = ()  # List sub vids to be displayed in this tab.
    relations_info = ()

    def entity_call(self, entity):
        if self.relations_info:
            relations_info = [RelationInfo(*rinfo) for rinfo in self.relations_info]
            urlparams = {'__redirectparams': 'tab=' + self.__regid__}
            button = add_relations_button(self._cw, entity, _('add'),
                                          *relations_info, **urlparams)
            if button is not None:
                # No button if user cannot add any relation.
                self.w(button)
                self.w(tags.div(klass='clearfix'))
        for vid in self.subvids:
            try:
                entity.view(vid, w=self.w, tabid=self.__regid__)
            except NoSelectableObject:
                # no data to display, skip view
                continue


class EditLinkView(EntityView):
    """Render entity as a link to the edition view"""
    __regid__ = 'saem.edit-link'

    def entity_call(self, entity, icon_info=True, editurlparams=None):
        editurlparams = editurlparams or {}
        links = []
        if icon_info:
            links.append(('icon-info', self._cw._('view'), entity.absolute_url()))
        if entity.cw_has_perm('update'):
            links.append(('icon-pencil', self._cw._('edit'),
                          entity.absolute_url(vid='edition', **editurlparams)))
        if entity.cw_has_perm('delete'):
            links.append(('icon-trash', self._cw._('delete'),
                          entity.absolute_url(vid='deleteconf', **editurlparams)))
        for csscls, label, href in links:
            self.w(u'<span title="{label}">{link}</span>'.format(
                label=label, link=tags.a(klass=csscls, href=href)))


urlrewrite.SchemaBasedRewriter.rules.append(
    (urlrewrite.rgx('/ark:/(.+)'),
     urlrewrite.build_rset(rql=r'Any X WHERE X ark %(text)s', rgxgroups=[('text', 1)]))
)


class SAEMRefRewriter(urlrewrite.SimpleReqRewriter):
    """URL rewriter for SAEM-Ref"""
    rules = [
        ('/oai', dict(vid='oai')),
        ('/ark', dict(vid='saem.ws.ark')),
        ('/archival-units', dict(vid='saem.ws.archival-units')),
        ('/sedalib', dict(rql='Any X WHERE X is SEDAArchiveUnit, '
                              'NOT X seda_archive_unit P',
                          vid='saem.sedalib')),
    ]


@monkeypatch(tabs.TabsMixin)
def active_tab(self, default):
    return self._cw.form.get('tab', domid(default))


class URLAttributeView(primary.URLAttributeView):
    """Overriden from cubicweb to handle case where ark is directly used in cwuri, and so absolute
    url should be generated from it.
    """
    def entity_call(self, entity, rtype, **kwargs):
        if rtype == 'cwuri':
            url = xml_escape(cwuri_url(entity))
        else:
            url = entity.printable_value(rtype)
        if url:
            self.w(u'<a href="%s">%s</a>' % (url, url))


class SAEMNoResultView(baseviews.NoResultView):
    """Enhanced "noresult" view with a nicer message and an invite to create
    an entity when permission is granted.
    """

    def call(self, **kwargs):
        vreg = self._cw.vreg
        try:
            etype = self.cw_rset.description[0][0]
        except IndexError:
            # rset probably comes from a user RQL query, fall back to default
            # behavior.
            return super(SAEMNoResultView, self).call(**kwargs)
        if etype not in vreg.schema.entities():
            # Or the first select query variable is not the entity type.
            return super(SAEMNoResultView, self).call(**kwargs)
        msg = self._cw._('No entity of type "{etype}" yet.').format(
            etype=self._cw.__(etype)
        )
        self.w(u'<div class="searchMessage">{}'.format(
            xml_escape(msg)))
        if vreg.schema[etype].has_perm(self._cw, 'add'):
            kwargs = ENTITY_CREATION_KWARGS.get(etype, {})
            url = vreg['etypes'].etype_class(etype).cw_create_url(self._cw, **kwargs)
            self.w(u' <a href="{url}">{invite}</a>'.format(
                url=url,
                invite=xml_escape(self._cw._('Create one?')),
            ))
        self.w(u'</div>')


class RestPathEvaluator(urlpublishing.RestPathEvaluator):

    def set_vid_for_rset(self, req, cls, rset):
        if rset.rowcount == 0:
            req.form['vid'] = 'noresult'
            return
        return super(RestPathEvaluator, self).set_vid_for_rset(req, cls, rset)


@ajaxcontroller.ajaxfunc(output_type='json')
def sort_relation_targets(self, rtype, eids):
    first_item = self._cw.entity_from_eid(eids[0]).cw_adapt_to('ISortable')
    collection = first_item.collection
    for item in collection:
        index = next(idx for idx, v in enumerate(eids) if v == text_type(item.eid))
        item.cw_set(index=index)


def registration_callback(vreg):
    from cubicweb.web.views import actions, cwuser, tableview, undohistory
    vreg.register_all(globals().values(), __name__, (
        RestPathEvaluator,
        SAEMHTMLPageFooter,
        SAEMNoResultView,
        URLAttributeView,
    ))
    vreg.register_and_replace(RestPathEvaluator, urlpublishing.RestPathEvaluator)
    vreg.register_and_replace(URLAttributeView, primary.URLAttributeView)
    vreg.register_and_replace(SAEMHTMLPageFooter, basetemplates.HTMLPageFooter)
    vreg.register_and_replace(SAEMNoResultView, baseviews.NoResultView)
    vreg.unregister(tableview.TableView)
    vreg.unregister(undohistory.UndoHistoryView)
    vreg.unregister(basecomponents.ApplicationName)
    # unregister some undesired actions
    vreg.unregister(actions.SelectAction)
    vreg.unregister(actions.CancelSelectAction)
    vreg.unregister(actions.ViewAction)
    vreg.unregister(actions.MultipleEditAction)
    vreg.unregister(actions.CopyAction)
    vreg.unregister(actions.AddNewAction)
    vreg.unregister(actions.ViewSameCWEType)
    vreg.unregister(actions.UserPreferencesAction)
    vreg.unregister(actions.ManageAction)
    vreg.unregister(actions.PoweredByAction)
    vreg.unregister(cwuser.UserPreferencesEntityAction)
    # global actions
    vreg.unregister(actions.SiteConfigurationAction)
    # facets
    from cubicweb_seda.views.facets import SimplifiedProfileFacet
    vreg.unregister(SimplifiedProfileFacet)

    for menu_entry in main_navigation_menu():
        vreg.register(menu_entry)



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref application package

Référenciel de Système d'Archivage Électronique Mutualisé
"""

from functools import partial

from six import text_type

from logilab.common.registry import objectify_predicate
from cubicweb import NoResultError

from cubicweb_compound import CompositeGraph


ConceptSchemeGraph = partial(CompositeGraph, skiprtypes=('generated', 'used'))


def cwuri_url(entity):
    """Return an absolute URL for entity's cwuri, handling case where ark is directly used, and so
    URL should be generated from it.
    """
    cwuri = entity.cwuri
    if cwuri.startswith('ark:'):
        cwuri = entity._cw.build_url(cwuri)
    return cwuri


def permanent_url(entity):
    """Return permanent URL for an entity: either ark based if entity has an ark, or <site url>/<eid>.
    """
    ark = getattr(entity, 'ark', None)
    if ark is not None:
        return entity._cw.build_url('ark:/' + ark)
    return entity._cw.build_url(text_type(entity.eid))


@objectify_predicate
def user_has_authority(cls, req, **kwargs):
    """Return 1 if the user is associated to an authority."""
    return len(req.user.authority)


@objectify_predicate
def user_has_naa(cls, req, **kwargs):
    """Return 1 if the user is associated to an authority with a NAA configured."""
    return 1 if req.user.naa is not None else 0


def includeme(config):
    config.include('.pviews')


def _metagen(cnx, config):
    from .sobjects import SAEMMetadataGenerator
    naa_what = getattr(config, 'naa_what', None)
    return SAEMMetadataGenerator(cnx, naa_what=naa_what)


def _massive_store_factory(cnx, config, **kwargs):
    from cubicweb.dataimport.massive_store import MassiveObjectStore
    return MassiveObjectStore(cnx, metagen=_metagen(cnx, config), **kwargs)


def _nohook_store_factory(cnx, config):
    from cubicweb.dataimport.stores import NoHookRQLObjectStore
    return NoHookRQLObjectStore(cnx, metagen=_metagen(cnx, config))


def ark_naa(cnx, who, what):
    """Get or create the ArkNameAssigningAuthority entity matchin "who" and
    "what" values.
    """
    if not isinstance(who, text_type):
        who = text_type(who, 'utf-8')
    try:
        return cnx.find('ArkNameAssigningAuthority', who=who, what=what).one()
    except NoResultError:
        return cnx.create_entity('ArkNameAssigningAuthority',
                                 who=who, what=what)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for Organization and its associated classes (mainly
OrganizationUnit and Agent)
"""

from cubicweb.predicates import is_instance
from cubicweb.entities import AnyEntity, fetch_config

from cubicweb_oaipmh.entities import ETypeOAISetSpec, RelatedEntityOAISetSpec

from . import oai


class Organization(AnyEntity):
    __regid__ = 'Organization'
    fetch_attrs, cw_fetch_order = fetch_config(('name', 'ark'))


class Agent(AnyEntity):
    __regid__ = 'Agent'
    fetch_attrs, cw_fetch_order = fetch_config(('name', 'ark'))


class OrganizationUnit(AnyEntity):
    __regid__ = 'OrganizationUnit'
    fetch_attrs, cw_fetch_order = fetch_config(('name', 'ark'))

    def has_role(self, role):
        """Return True if the unit has the archival role `role` else False"""
        return any(archival_role for archival_role in self.archival_role
                   if archival_role.name == role)


class ArchivalRole(AnyEntity):
    __regid__ = 'ArchivalRole'
    fetch_attrs, cw_fetch_order = fetch_config(('name',))


ORG_METADATA_FORMATS = {'rdf': (oai.RDF_METADATAFORMAT, 'primary.rdf')}


class OrganizationOAIPMHRecordAdapter(oai.ArkOAIPMHRecordAdapter):
    """OAI-PMH adapter for Organization entity type."""
    __select__ = (oai.ArkOAIPMHRecordAdapter.__select__
                  & is_instance('Organization'))
    metadata_formats = ORG_METADATA_FORMATS.copy()

    @classmethod
    def set_definition(cls):
        return ETypeOAISetSpec(
            'Organization', identifier_attribute=cls.identifier_attribute)


class OrganizationUnitOAIPMHRecordAdapter(oai.OAIPMHActiveRecordAdapter):
    """OAI-PMH adapter for OrganizationUnit entity type."""
    __select__ = (oai.OAIPMHActiveRecordAdapter.__select__
                  & is_instance('OrganizationUnit'))
    metadata_formats = ORG_METADATA_FORMATS.copy()

    @classmethod
    def set_definition(cls):
        specifier = oai.PublicETypeOAISetSpec(
            'OrganizationUnit',
            identifier_attribute=cls.identifier_attribute)
        specifier['role'] = RelatedEntityOAISetSpec(
            'archival_role', 'ArchivalRole', 'name',
            description=u'An organization unit with {0} archival role')
        return specifier


class DeletedOrganizationUnitOAIPMHRecordAdapter(oai.OAIPMHDeletedRecordAdapter):
    """OAI-PMH adapter for deleted OrganizationUnit."""
    __select__ = (oai.OAIPMHDeletedRecordAdapter.__select__
                  & is_instance('OrganizationUnit'))
    metadata_formats = ORG_METADATA_FORMATS.copy()


class AgentOAIPMHRecordAdapter(oai.OAIPMHActiveRecordAdapter):
    """OAI-PMH adapter for Agent entity type."""
    __select__ = (oai.OAIPMHActiveRecordAdapter.__select__
                  & is_instance('Agent'))
    metadata_formats = ORG_METADATA_FORMATS.copy()

    @classmethod
    def set_definition(cls):
        return oai.PublicETypeOAISetSpec(
            'Agent',
            identifier_attribute=cls.identifier_attribute)


class DeletedAgentOAIPMHRecordAdapter(oai.OAIPMHDeletedRecordAdapter):
    """OAI-PMH adapter for deleted Agent."""
    __select__ = (oai.OAIPMHDeletedRecordAdapter.__select__
                  & is_instance('Agent'))
    metadata_formats = ORG_METADATA_FORMATS.copy()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes"""

from logilab.common.decorators import monkeypatch

from cubicweb import neg_role
from cubicweb.view import Adapter, EntityAdapter
from cubicweb.predicates import match_kwargs, relation_possible, is_instance
from cubicweb.entities import AnyEntity, fetch_config, authobjs, lib

from .. import ark


class ARKIdentifierGeneratorMixin(object):
    __regid__ = 'IARKGenerator'
    __select__ = match_kwargs('naa_what')

    def generate_ark(self):
        """Return a new ARK identifier as unicode"""
        return u'{0}/{1}'.format(self.cw_extra_kwargs['naa_what'], self.assign_name())


class ARKIdentifierGenerator(ARKIdentifierGeneratorMixin, Adapter):
    """Adapter for ARK unique identifier generation"""

    def assign_name(self):
        """Assign and return a new name part of the ARK identifier"""
        naan = int(self.cw_extra_kwargs['naa_what'])
        return ark.generate_ark(self._cw, naan)


class QualifiedARKIdentifierGenerator(ARKIdentifierGeneratorMixin, Adapter):
    """Adapter for ARK identifier generation using a qualifier from an
    existing "base" ARK identifier.
    """
    __abstract__ = True

    @property
    def parent_entity(self):
        raise NotImplementedError

    def assign_name(self):
        match = ark.match(self.parent_entity.ark)
        if not match:
            raise ValueError(
                "ARK identifier for parent entity #%d looks malformattted: %s"
                % (self.parent_entity.eid, self.parent_entity.ark)
            )
        naan = int(self.cw_extra_kwargs['naa_what'])
        # Sanity check to make sure we're not producing ARK identifiers with
        # different NAAN for entities that are supposed to be parent of each
        # others.
        if int(match.group('naan')) != naan:
            raise ValueError(
                "NAAN of parent entity ark:/%s does not match the value of "
                "'naa_what' parameter: %s"
                % (self.parent_entity.ark, naan)
            )
        name = match.group('name')
        qualifier = ark.generate_qualified_ark(self._cw, naan, name)
        return u'/'.join([name, qualifier])


class ConceptARKIdentifierGenerator(QualifiedARKIdentifierGenerator):
    __select__ = (
        QualifiedARKIdentifierGenerator.__select__
        & match_kwargs('in_scheme')
    )

    @property
    def parent_entity(self):
        scheme = self.cw_extra_kwargs['in_scheme']
        if hasattr(scheme, 'eid'):
            return scheme
        return self._cw.entity_from_eid(scheme)


class OUARKIdentifierGenerator(QualifiedARKIdentifierGenerator):
    __select__ = (
        QualifiedARKIdentifierGenerator.__select__
        & match_kwargs('authority')
    )

    @property
    def parent_entity(self):
        authority = self.cw_extra_kwargs['authority']
        if hasattr(authority, 'eid'):
            return authority
        return self._cw.entity_from_eid(authority)


def archive_unit_parent(parent):
    if parent.cw_etype == 'SEDAArchiveTransfer':
        return parent
    elif parent.cw_etype == 'SEDASeqAltArchiveUnitArchiveUnitRefIdManagement':
        alt = parent.reverse_seda_seq_alt_archive_unit_archive_unit_ref_id_management[0]
        au = alt.reverse_seda_alt_archive_unit_archive_unit_ref_id[0]
        assert au.cw_etype == 'SEDAArchiveUnit', au
        if au.seda_archive_unit:
            parent = au.seda_archive_unit[0]
            return archive_unit_parent(parent)
        else:
            return au
    else:
        raise ValueError('{} should be target of seda_archive_unit relation'.format(parent))


class SEDAArchiveUnitARKIdentifierGenerator(QualifiedARKIdentifierGenerator):
    __select__ = (
        QualifiedARKIdentifierGenerator.__select__
        & match_kwargs('seda_archive_unit')
    )

    @property
    def parent_entity(self):
        target = self.cw_extra_kwargs['seda_archive_unit']
        if not hasattr(target, 'eid'):
            target = self._cw.entity_from_eid(target)
        return archive_unit_parent(target)


class ArkNAALocator(EntityAdapter):
    """Adapter responsible to retrieve the proper ARK Name Assigning Authority depending on the
    entity type
    """
    __abstract__ = True
    __regid__ = 'IArkNAALocator'

    def naa_what(self):
        """Return the ARK NameAssigningAuthority entity or None if not specified"""
        raise NotImplementedError()


def direct_naa_what(entity):
    # entity is usually not yet created, since ark has to be generated before entity creation
    if 'ark_naa' in getattr(entity, 'cw_edited', {}):
        return entity._cw.entity_from_eid(entity.cw_edited['ark_naa']).what
    elif entity.ark_naa:
        return entity.ark_naa[0].what
    return None


class DirectArkNAALocator(ArkNAALocator):
    """Return NAA specified through the ark_naa relation"""
    __select__ = relation_possible('ark_naa')

    def naa_what(self):
        return direct_naa_what(self.entity)


class AgentArkNAALocator(ArkNAALocator):
    """Return NAA specified through the authority to which the agent belong"""
    __select__ = is_instance('Agent', 'OrganizationUnit')

    def naa_what(self):
        # entity is usually not yet created, since ark has to be generated before entity creation
        if 'authority' in getattr(self.entity, 'cw_edited', {}):
            authority = self._cw.entity_from_eid(self.entity.cw_edited['authority'])
            if authority.ark_naa:
                return authority.ark_naa[0].what
        elif self.entity.authority and self.authority[0].ark_naa:
            return self.authority[0].ark_naa[0].what
        return None


class ConceptArkNAALocator(ArkNAALocator):
    """Return NAA for Concept, through the concept scheme which it belongs to."""
    __select__ = is_instance('Concept')

    def naa_what(self):
        # entity is usually not yet created, since ark has to be generated before entity creation
        if 'in_scheme' in getattr(self.entity, 'cw_edited', {}):
            scheme = self._cw.entity_from_eid(self.entity.cw_edited['in_scheme'])
        elif self.entity.in_scheme:
            scheme = self.entity.in_scheme[0]
        else:
            return None
        return scheme.cw_adapt_to(self.__regid__).naa_what()


class SEDAArchiveUnitArkNAALocator(ArkNAALocator):
    """Return NAA for a SEDAArchiveUnit, either direct or through the
    SEDAArchiveTransfer it belongs to.
    """
    __select__ = is_instance('SEDAArchiveUnit')

    def naa_what(self):
        edited = getattr(self.entity, 'cw_edited', {})
        # entity is usually not yet created, since ark has to be generated before entity creation
        if 'seda_archive_unit' in edited or self.entity.seda_archive_unit:
            if 'seda_archive_unit' in edited:
                target = self._cw.entity_from_eid(edited['seda_archive_unit'])
            else:
                target = self.entity.seda_archive_unit[0]
            target = archive_unit_parent(target)
            return target.cw_adapt_to(self.__regid__).naa_what()
        else:
            return direct_naa_what(self.entity)


class ExternalUri(AnyEntity):
    __regid__ = 'ExternalUri'
    fetch_attrs, cw_fetch_order = fetch_config(('uri',))


class EmailAddress(lib.EmailAddress):
    fetch_attrs, cw_fetch_order = fetch_config(['address'])

    def dc_title(self):
        return self.display_address()


@monkeypatch(authobjs.CWUser, methodname='naa')
@property
def naa(self):
    if self.authority and self.authority[0].ark_naa:
        return self.authority[0].ark_naa[0]
    return None


class ISortable(EntityAdapter):
    __regid__ = 'ISortable'
    __abstract__ = True

    role = 'subject'
    rtype = None

    @property
    def parent(self):
        return self.entity.related(self.rtype, self.role, entities=True)[0]

    @property
    def collection(self):
        return self.parent.related(self.rtype, neg_role(self.role), entities=True)


class AuthorityRecordFunctionSortable(ISortable):
    __select__ = is_instance('AgentFunction')

    rtype = 'function_agent'


class AuthorityRecordMandateSortable(ISortable):
    __select__ = is_instance('Mandate')

    rtype = 'mandate_agent'



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Auto-completion in search widgets."""

from collections import defaultdict

from logilab.mtconverter import xml_escape
from rql import RQLException

from cubicweb.uilib import domid
from cubicweb.utils import json_dumps, make_uid
from cubicweb.view import StartupView
from cubicweb.web import facet as facetbase
from cubicweb.web.views.boxes import SearchBox
from cubicweb.web.views.facets import HasTextFacet

__docformat__ = "restructuredtext en"

# XXX The words selection should be done in RQL ::
#     Any X HAVING W=WORDS(%(q)s)
#     or
#     Any W, SIMILARITY(W) HAVING W=WORDS(%(q)s)
#
# However, with WORDS class declared as above::
#
#     class WORDS(FunctionDescr):
#         supported_backends = ('postgres',)
#         rtype = 'String'
#
# the resulting sql for Any X HAVING W=WORDS('toto') states :
#
#       SELECT _W.eid
#       FROM entities AS _W
#       WHERE _W.eid=WORDS(%(140390076097112)s) {'140390076097112': u'toto'}
#
# Some more modifications are needed in `server/sources/rql2sql.py`,
# which we dont't want to do for instance.


def words_matching_term(req, term, previous_words='', base_rql=''):
    """Return a similarity ordered list of (word, etype) matching the current term, considering
    previously typed words and optional base rql in the case of facetted search.
    """
    assert term
    if not base_rql:
        base_rql = 'Any X WHERE X has_text %(text)s, X is ET, ET name %(etype)s'
    else:
        base_rql += ', X has_text %(text)s, X is ET, ET name %(etype)s'
    # the sqlgenerator introduces the searched text in the generated SQL and not in the query
    # arguments. As we want to hack this later, introduces an unexpected placeholder that will be
    # replaced.
    text_place_holder = 'xyz' * 10
    args = {'text': text_place_holder, 'etype': 'whatever'}
    # generate SQL from RQL to join with the above query
    try:
        union = req.vreg.parse(req, base_rql)
    except RQLException:
        return []
    req.vreg.rqlhelper.simplify(union)
    repository = req.cnx.repo
    plan = repository.querier.plan_factory(union, args, req.cnx)
    plan.preprocess(union)
    base_sql, qargs, _ = repository.system_source._rql_sqlgen.generate(union, args)
    # we don't care of original ordering, remove it
    base_sql = base_sql.split('ORDER BY')[0]
    # replace plain-text search terms and etype substitution by element from the parent query
    if previous_words:
        searched = repository.system_source.dbhelper._fti_query_to_tsquery_words(previous_words)
        searched = "'{0}&' || normalized_word".format(searched)
    else:
        searched = "normalized_word"
    base_sql = base_sql.replace("'{0}'".format(text_place_holder), searched)
    base_sql = base_sql.replace("%(etype)s", 'etype')
    qargs['q'] = term
    sql = '''SELECT word, etype
FROM words
WHERE similarity(word, %(q)s) > 0.4 AND length(word) > length(%(q)s)-1
AND EXISTS({base_sql})
ORDER BY similarity(word, %(q)s) DESC, word, etype;'''.format(base_sql=base_sql)
    return req.cnx.system_sql(sql, qargs).fetchall()


class AutocompleteSearchView(StartupView):
    """The principal view for autocomplete search
    """
    __regid__ = 'search-autocomplete'
    templatable = False
    binary = True

    def call(self):
        base_rql = self._cw.form['facetrql']
        # string containing words typed before the current one, to later ensure we provide
        # consistent results
        previous_words = ' '.join(self._cw.form['all'].split()[:-1])
        matching_words = words_matching_term(self._cw, self._cw.form['term'],
                                             previous_words, base_rql)
        processed = defaultdict(list)
        words = []
        for word, etype in matching_words:
            if word not in processed:
                words.append(word)
            processed[word].append(self._cw._(etype))
        results = []
        for word in words:
            labels = processed[word]
            results.append({'label': '%s : %s' % (word, ', '.join(sorted(labels))),
                            'value': word})
        self.w(json_dumps(results).encode('utf-8'))


class AutocompleteSearchBox(SearchBox):
    formdef = '''
<form action="%(action)s" id="search_box" class="navbar-form" role="search">
  <input id="norql" autocomplete="off" type="text" accesskey="q" tabindex="%(tabindex1)s"
    title="search text" value="%(value)s" name="rql"
    class="search-query form-control" placeholder="%(searchlabel)s"/>
  <input type="hidden" name="__fromsearchbox" value="1" />
  <input type="hidden" name="subvid" value="tsearch" />
</form>
'''

    def render_body(self, w):
        self._cw.add_js(('jquery.ui.js', 'cubicweb.autocompletesearch.js'))
        self._cw.add_css(('jquery.ui.css',))
        url = self._cw.build_url('view', vid='search-autocomplete')
        minlength = 2
        limit = 20
        self._cw.add_onload('cw.search_autocomplete($("#norql"), "%s", %s, %s);'
                            % (url, minlength, limit))
        super(AutocompleteSearchBox, self).render_body(w)


class AutocompleteSearchWidget(facetbase.FacetStringWidget):

    def _render(self):
        """ has-hext facet widget with autocomplete"""
        req = self.facet._cw
        facetid = domid(make_uid(self.facet.__regid__))
        input_id = xml_escape('id-%s' % facetid)
        req.add_js(('jquery.ui.js', 'cubicweb.autocompletesearch.js'))
        req.add_css(('jquery.ui.css', ))
        url = req.build_url('view', vid='search-autocomplete')
        minlength = 2
        limit = 20
        req.add_onload('cw.search_autocomplete($("#%s"), "%s", %s, %s);' % (
            input_id, url, minlength, limit))
        w = self.w
        title = xml_escape(self.facet.title)
        w(u'<div id="%s" class="facet">\n' % facetid)
        cssclass = 'facetTitle'
        if self.facet.allow_hide:
            cssclass += ' hideFacetBody'
        w(u'<div class="%s" cubicweb:facetName="%s">%s</div>\n' %
            (cssclass, xml_escape(self.facet.__regid__), title))
        cssclass = 'facetBody'
        if not self.facet.start_unfolded:
            cssclass += ' hidden'
        w(u'<div class="%s">\n' % cssclass)
        w((u'<input autocomplete="off" id="%s" name="%s" '
            'type="text" value="%s" data-facet="1"/>\n') % (
            input_id, xml_escape(self.facet.__regid__),
            self.value or u''))
        w(u'</div>\n')
        w(u'</div>\n')


def registration_callback(vreg):
    try:
        import psycopg2
    except ImportError:
        psycopg2 = None
    if psycopg2:
        vreg.register_all(globals().values(), __name__, [])
        vreg.unregister(SearchBox)
        HasTextFacet.wdgclass = AutocompleteSearchWidget



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-ctl plugin customizing data import commands."""
from __future__ import print_function

import sys

from six import PY2

from logilab.common.decorators import monkeypatch

from cubicweb import MultipleResultsError, NoResultError
from cubicweb.toolsutils import underline_title
from cubicweb.utils import admincnx

from cubicweb_skos import ccplugin as skos
from cubicweb_eac import ccplugin as eac

from . import _massive_store_factory, _nohook_store_factory


eac.ImportEacData.options = (
    ("authority", {
        'short': 'n', 'type': 'string',
        'default': False,
        'help': ('Name of the reference authority to use while importing data. This authority '
                 'should be linked to an NAA that will be used to attribute ARK identifiers. '
                 'If not specified, a single one is expected to be found in the database.'),
    }),
)


def _skos_drop_rql_store():
    """Remove "rql" store from ImportSkosData command.

    "rql" store is not supported in saem_ref because it cannot handle a
    metadata generator which is needed in saem_ref to generate ARK
    identifiers.
    """
    for name, value in skos.ImportSkosData.options:
        if name == 'cw-store':
            value['choices'] = tuple(c for c in value['choices'] if c != 'rql')
            assert 'nohook' in value['choices'], value['choices']
            value['default'] = 'nohook'
            break
    else:
        raise AssertionError(
            'Could not find "cw-store" option in ImportSkosData ccplugin command')
    del skos.ImportSkosData.cw_store_factories['rql']


_skos_drop_rql_store()
del _skos_drop_rql_store


def _skos_reword_scheme_option():
    """Change "--scheme" option help of skos-import command to match SAEM
    semantics (which relies on ARK instead of bare URI).
    """
    for name, value in skos.ImportSkosData.options:
        if name == 'scheme':
            value['help'] = (
                'ARK identifier of an existing concept scheme to import concepts in '
                '(only relevant for LCSV import format)'
            )
            break
    else:
        raise AssertionError(
            'Could not find "scheme" option in ImportSkosData ccplugin command')


_skos_reword_scheme_option()
del _skos_reword_scheme_option


def _enc(string):
    if PY2 and not isinstance(string, str):
        string = string.encode('utf-8')
    return string


@monkeypatch(eac.ImportEacData)
def run(self, args):
    appid = args[0]
    repo = None
    try:
        with admincnx(appid) as cnx:
            repo = cnx.repo
            org_name = self.config.authority
            if org_name:
                try:
                    org = cnx.find('Organization', name=org_name).one()
                except NoResultError:
                    print(u'ERROR: there are no authority named "{}"'.format(org_name))
                    sys.exit(1)
            else:
                org_rset = cnx.find('Organization')
                try:
                    org = org_rset.one()
                except MultipleResultsError:
                    print(u'ERROR: there are several authorities, choose the one to use using '
                          u'--authority option. It should be one of {}'.format(
                              u', '.join(_enc(org.name) for org in org_rset.entities())))
                    sys.exit(1)
            if not org.ark_naa:
                print(u'ERROR: authority {} is not associated to a ARK naming authority. Choose '
                      u'another organization or associate it to a ARK naming authority first.'
                      .format(_enc(org.name)))
                sys.exit(1)

            print(u'\n%s' % underline_title(u'Importing EAC files'))
            if cnx.repo.system_source.dbdriver == 'postgres':
                store = _massive_store_factory(cnx, self.config, eids_seq_range=100)
                store.metagen.naa_what = org.ark_naa[0].what
            else:
                store = _nohook_store_factory(cnx, self.config)
            eac.eac_import_files(cnx, args[1:], authority=org, store=store)
    finally:
        if repo is not None:
            repo.shutdown()


_orig_run = skos.ImportSkosData.run


@monkeypatch(skos.ImportSkosData)  # noqa: F811
def run(self, args):
    if self.get('format') == 'lcsv':
        scheme_ark = self.get('scheme')
        if not scheme_ark:
            print(u'command failed: --scheme option is required for LCSV import')
            sys.exit(1)
        scheme_ark = scheme_ark.strip()
        if PY2:
            scheme_ark = scheme_ark.decode('utf-8')
        appid = args[0]
        connection = self.get_cnx(appid)
        try:
            with connection as cnx:
                if scheme_ark.startswith(u'ark:/'):
                    scheme_ark = scheme_ark[len(u'ark:/'):]
                rset = cnx.find('ConceptScheme', ark=scheme_ark)
                if not rset:
                    print(u'command failed: no concept scheme found matching "%s"'
                          % scheme_ark)
                    raise Exception
                scheme = rset.one()
                # cubicweb-skos's command need a cwuri for --scheme.
                setattr(self.config, 'scheme', scheme.cwuri)
                # Per schema, ark_naa is optional for ConceptScheme.
                if not scheme.ark_naa:
                    print(u'command failed: specified concept scheme "%s" must have an ARK NAA set'
                          % scheme_ark)
                    raise Exception
                naa_what = scheme.ark_naa[0].what
                print(u'Importing concepts in %s' % scheme.absolute_url())
        except Exception:
            connection.repo.shutdown()
            sys.exit(1)
        # Set a "config" attribute which will be retrieved from metadata
        # generator defined in __init__.py.
        setattr(self.config, 'naa_what', naa_what)
    return _orig_run(self, args)



# copyright 2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref utilities for ARK identifiers"""

import re

from six import text_type


ARK_MAXIT = 20
ARK_PREFIX = 'rf'
ARK_CONTROLCHAR = 'g'
ARK_NAME_LENGTH = 10
ARK_QUALIFIER_LENGTH = 10

ARK_RGX = re.compile(
    (r'^(ark:/)?(?P<naan>\d+)'
     r'/(?P<name>{prefix}\w{{{length}}}{controlchar})'
     r'(/(?P<qualifier>.{{{qualifier_length}}}))?$').format(
        prefix=ARK_PREFIX,
        length=ARK_NAME_LENGTH - len(ARK_PREFIX) - len(ARK_CONTROLCHAR),
        controlchar=ARK_CONTROLCHAR,
        qualifier_length=ARK_QUALIFIER_LENGTH,
    ),
)

# Regexp for "external" ARK identifiers (i.e. with no length, prefix or
# control character).
EXT_ARK_RGX = re.compile(
    r'^(ark:/)?(?P<naan>\d+)/(?P<name>\w+)(/(?P<qualifier>\w+))?$',
)


def match(string, external=False, **kwargs):
    if external:
        return EXT_ARK_RGX.match(string, **kwargs)
    else:
        return ARK_RGX.match(string, **kwargs)


def generate_ark(cnx, naan):
    """Insert a record in "ark" table with "naan" value, a "name" generated
    and an empty qualifier.

    Return the generated name.
    """
    cu = cnx.system_sql(
        'select * from gen_ark(%s, %s, %s, %s, %s);',
        (naan, ARK_NAME_LENGTH, ARK_PREFIX, ARK_CONTROLCHAR, ARK_MAXIT),
    )
    name, = cu.fetchone()
    return name


def generate_qualified_ark(cnx, naan, name):
    """Insert a record in "ark" table with "naan" and "name" values and a
    "qualifier" generated.

    Return the generated qualifier.
    """
    cu = cnx.system_sql(
        'select * from gen_qualified_ark(%s, %s, %s, %s);',
        (naan, name, ARK_QUALIFIER_LENGTH, ARK_MAXIT),
    )
    qualifier, = cu.fetchone()
    return qualifier


def insert_ark(cnx, naan, name, qualifier=None):
    """Insert a record in "ark" table from specified values."""
    if qualifier is None:
        qs = 'INSERT INTO ark VALUES (%s, %s) RETURNING naan, name, qualifier;'
        values = (naan, name)
    else:
        qs = 'INSERT INTO ark VALUES (%s, %s, %s) RETURNING naan, name, qualifier;'
        values = (naan, name, qualifier)
    (naan, name, qualifier), = cnx.system_sql(qs, values).fetchall()
    ark = u'/'.join([text_type(naan), name])
    if qualifier:
        ark += u'/' + qualifier
    return ark




from __future__ import print_function

print('create source for data.culture.fr')
create_entity('CWSource', name=u'data.culture.fr', type=u'datafeed', parser=u'rdf.skos',
              url=u'http://data.culture.fr/thesaurus/data/ark:/67717/Matiere?includeSchemes=true\n'
              'http://data.culture.fr/thesaurus/data/ark:/67717/T2?includeSchemes=true\n'
              'http://data.culture.fr/thesaurus/data/ark:/67717/T4?includeSchemes=true\n'
              'http://data.culture.fr/thesaurus/data/ark:/67717/T3?includeSchemes=true\n',
              config=u'synchronize=no\nuse-cwuri-as-url=no')
commit()

print('importing data.culture.fr vocabularies (takes some time)')
with repo.internal_cnx() as cnx:
    dfsource = repo.sources_by_uri['data.culture.fr']
    stats = dfsource.pull_data(cnx, force=True, raise_on_error=True)



# coding: utf-8
# pylint: disable=W0622
"""cubicweb-saem-ref application packaging information"""

distname = 'cubicweb-saem-ref'
modname = 'cubicweb_saem_ref'

numversion = (0, 25, 1)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = "Référenciel de Système d'Archivage Électronique Mutualisé"
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb[pyramid]': '>= 3.26.1',
    'cubicweb-squareui': '>= 1.1.2',
    'cubicweb-eac': '>= 0.8.4, < 0.9.0',
    'cubicweb-seda': '>= 0.17.2, < 0.18.0',
    'cubicweb-compound': '>= 0.7.0, < 0.8.0',
    'cubicweb-prov': '>= 0.4.0, < 0.5.0',
    'cubicweb-oaipmh': '>= 0.6.0, < 0.7.0',
    'cubicweb-relationwidget': '>= 0.5.1, < 0.6.0',
    'cubicweb-skos': '>= 1.6.0',
    'cubicweb-vtimeline': '>= 0.6.0',
    'cubicweb-signedrequest': None,
    'isodate': None,
    'python-dateutil': None,
    'pytz': None,
    'psycopg2': '< 2.8',
    'rdflib': '>= 4.1',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from lxml import etree

from cubicweb.devtools import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC

import testutils


def setUpModule():
    testutils.startpgcluster(__file__)


class EACExportTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_associated_with(self):
        with self.admin_access.client_cnx() as cnx:
            record = testutils.authority_record(cnx, u'My record')
            cnx.create_entity('Activity', agent=u'007', start=datetime.utcnow(),
                              generated=record)
            cnx.commit()
            eac_xml = record.cw_adapt_to('EAC-CPF').dump().decode('utf-8')
            self.assertIn('<agent>admin</agent>', eac_xml)
            self.assertIn('<agent>007</agent>', eac_xml)

    def test_export_record_id(self):
        with self.admin_access.cnx() as cnx:
            record = testutils.authority_record(cnx, u'My record',
                                                record_id=u'My RID')
            cnx.create_entity('EACOtherRecordId', value=u'Other RID',
                              eac_other_record_id_of=record)
            cnx.commit()
            record = cnx.entity_from_eid(record.eid)
            xml = record.cw_adapt_to('EAC-CPF').dump()
            schema = etree.XMLSchema(etree.parse(self.datapath('cpf.xsd')))
            schema.assert_(etree.fromstring(xml))

            xml = xml.decode('utf-8')
            self.assertIn('<recordId>{0}</recordId>'.format(record.ark.replace('/', '-')), xml)
            self.assertIn('<otherRecordId>My RID</otherRecordId>', xml)
            self.assertIn('<otherRecordId>Other RID</otherRecordId>', xml)


if __name__ == '__main__':
    import unittest
    unittest.main()
